import json
import riak.pb.messages

from riak.util import decode_index_value, bytes_to_str
from riak.client.index_page import CONTINUATION
from riak.transports.pbc.codec import RiakPbcCodec
from six import PY2


class RiakPbcStream(object):
    """
    Used internally by RiakPbcTransport to implement streaming
    operations. Implements the iterator interface.
    """

    _expect = None

    def __init__(self, transport):
        self.finished = False
        self.transport = transport
        self.resource = None

    def __iter__(self):
        return self

    def next(self):
        if self.finished:
            raise StopIteration

        try:
            msg_code, resp = self.transport._recv_msg(expect=self._expect)
        except:
            self.finished = True
            raise

        if self._is_done(resp):
            self.finished = True

        return resp

    def __next__(self):
        # Python 3.x Version
        return self.next()

    def _is_done(self, response):
        # This could break if new messages don't name the field the
        # same thing.
        return response.done

    def attach(self, resource):
        self.resource = resource

    def close(self):
        # We have to drain the socket to make sure that we don't get
        # weird responses when some other request comes after a
        # failed/prematurely-terminated one.
        try:
            while self.next():
                pass
        except StopIteration:
            pass
        self.resource.release()


class RiakPbcKeyStream(RiakPbcStream):
    """
    Used internally by RiakPbcTransport to implement key-list streams.
    """

    _expect = riak.pb.messages.MSG_CODE_LIST_KEYS_RESP

    def next(self):
        response = super(RiakPbcKeyStream, self).next()

        if response.done and len(response.keys) is 0:
            raise StopIteration

        return response.keys

    def __next__(self):
        # Python 3.x Version
        return self.next()


class RiakPbcMapredStream(RiakPbcStream):
    """
    Used internally by RiakPbcTransport to implement MapReduce
    streams.
    """

    _expect = riak.pb.messages.MSG_CODE_MAP_RED_RESP

    def next(self):
        response = super(RiakPbcMapredStream, self).next()

        if response.done and not response.HasField('response'):
            raise StopIteration

        return response.phase, json.loads(bytes_to_str(response.response))

    def __next__(self):
        # Python 3.x Version
        return self.next()


class RiakPbcBucketStream(RiakPbcStream):
    """
    Used internally by RiakPbcTransport to implement key-list streams.
    """

    _expect = riak.pb.messages.MSG_CODE_LIST_BUCKETS_RESP

    def next(self):
        response = super(RiakPbcBucketStream, self).next()

        if response.done and len(response.buckets) is 0:
            raise StopIteration

        return response.buckets

    def __next__(self):
        # Python 3.x Version
        return self.next()


class RiakPbcIndexStream(RiakPbcStream):
    """
    Used internally by RiakPbcTransport to implement Secondary Index
    streams.
    """

    _expect = riak.pb.messages.MSG_CODE_INDEX_RESP

    def __init__(self, transport, index, return_terms=False):
        super(RiakPbcIndexStream, self).__init__(transport)
        self.index = index
        self.return_terms = return_terms

    def next(self):
        response = super(RiakPbcIndexStream, self).next()

        if response.done and not (response.keys or
                                  response.results or
                                  response.continuation):
            raise StopIteration

        if self.return_terms and response.results:
            return [(decode_index_value(self.index, r.key),
                     bytes_to_str(r.value))
                    for r in response.results]
        elif response.keys:
            if PY2:
                return response.keys[:]
            else:
                return [bytes_to_str(key) for key in response.keys]
        elif response.continuation:
            return CONTINUATION(bytes_to_str(response.continuation))

    def __next__(self):
        # Python 3.x Version
        return self.next()


class RiakPbcTsKeyStream(RiakPbcStream, RiakPbcCodec):
    """
    Used internally by RiakPbcTransport to implement key-list streams.
    """

    _expect = riak.pb.messages.MSG_CODE_TS_LIST_KEYS_RESP

    def next(self):
        response = super(RiakPbcTsKeyStream, self).next()

        if response.done and len(response.keys) is 0:
            raise StopIteration

        keys = []
        for tsrow in response.keys:
            keys.append(self._decode_timeseries_row(tsrow))

        return keys

    def __next__(self):
        # Python 3.x Version
        return self.next()



# -*- coding: utf-8 -*-
from six import PY2
import platform
from threading import Thread, currentThread
from riak.transports.pool import Pool, BadResource
from random import SystemRandom
from time import sleep
from riak.tests import RUN_POOL
from riak.tests.comparison import Comparison

if platform.python_version() < '2.7':
    unittest = __import__('unittest2')
else:
    import unittest

if PY2:
    from Queue import Queue
else:
    from queue import Queue


class SimplePool(Pool):
    def __init__(self):
        self.count = 0
        Pool.__init__(self)

    def create_resource(self):
        self.count += 1
        return [self.count]

    def destroy_resource(self, resource):
        del resource[:]


class EmptyListPool(Pool):
    def create_resource(self):
        return []


@unittest.skipUnless(RUN_POOL, 'RUN_POOL is 0')
class PoolTest(unittest.TestCase, Comparison):

    def test_yields_new_object_when_empty(self):
        """
        The pool should create new resources as needed.
        """
        pool = SimplePool()
        with pool.transaction() as element:
            self.assertEqual([1], element)

    def test_yields_same_object_in_serial_access(self):
        """
        The pool should reuse resources that already exist, when used
        serially.
        """
        pool = SimplePool()

        with pool.transaction() as element:
            self.assertEqual([1], element)
            element.append(2)

        with pool.transaction() as element2:
            self.assertEqual(1, len(pool.resources))
            self.assertEqual([1, 2], element2)

        self.assertEqual(1, len(pool.resources))

    def test_reentrance(self):
        """
        The pool should be re-entrant, that is, yield new resources
        while one is already claimed in the same code path.
        """
        pool = SimplePool()
        with pool.transaction() as first:
            self.assertEqual([1], first)
            with pool.transaction() as second:
                self.assertEqual([2], second)
                with pool.transaction() as third:
                    self.assertEqual([3], third)

    def test_unlocks_when_exception_raised(self):
        """
        The pool should unlock all resources that were previously
        claimed when an exception occurs.
        """
        pool = SimplePool()
        try:
            with pool.transaction():
                with pool.transaction():
                    raise RuntimeError
        except:
            self.assertEqual(2, len(pool.resources))
            for e in pool.resources:
                self.assertFalse(e.claimed)

    def test_removes_bad_resource(self):
        """
        The pool should remove resources that are considered bad by
        user code throwing a BadResource exception.
        """
        pool = SimplePool()
        with pool.transaction() as resource:
            self.assertEqual([1], resource)
            resource.append(2)
        try:
            with pool.transaction():
                raise BadResource
        except BadResource:
            self.assertEqual(0, len(pool.resources))
            with pool.transaction() as goodie:
                self.assertEqual([2], goodie)

    def test_filter_skips_unmatching_resources(self):
        """
        The _filter parameter should cause the pool to yield the first
        unclaimed resource that passes the filter.
        """
        def filtereven(numlist):
            return numlist[0] % 2 == 0

        pool = SimplePool()
        with pool.transaction():
            with pool.transaction():
                pass

        with pool.transaction(_filter=filtereven) as f:
            self.assertEqual([2], f)

    def test_requires_filter_to_be_callable(self):
        """
        The _filter parameter should be required to be a callable, or
        None.
        """
        badfilter = 'foo'
        pool = SimplePool()

        with self.assertRaises(TypeError):
            with pool.transaction(_filter=badfilter):
                pass

    def test_yields_default_when_empty(self):
        """
        The pool should yield the given default when no existing
        resources are free.
        """
        pool = SimplePool()
        with pool.transaction(default='default') as x:
            self.assertEqual('default', x)

    def test_manual_release(self):
        """
        The pool should allow resources to be acquired and released
        manually, without giving them out twice.
        """
        pool = SimplePool()
        a = pool.acquire()
        self.assertEqual([1], a.object)
        with pool.transaction() as b:
            self.assertEqual([2], b)
        with pool.transaction() as c:
            self.assertEqual([2], c)
        pool.release(a)
        with pool.transaction() as d:
            self.assertEqual([1], d)
        e = pool.acquire()
        with pool.transaction() as f:
            self.assertEqual([2], f)
        e.release()
        with pool.transaction() as g:
            self.assertEqual([1], g)

    def test_thread_safety(self):
        """
        The pool should allocate n objects for n concurrent operations.
        """
        n = 10
        pool = EmptyListPool()
        readyq = Queue()
        finishq = Queue()
        threads = []

        def _run():
            with pool.transaction() as resource:
                readyq.put(1)
                resource.append(currentThread())
                finishq.get(True)
                finishq.task_done()

        for i in range(n):
            th = Thread(target=_run)
            threads.append(th)
            th.start()

        for i in range(n):
            readyq.get()
            readyq.task_done()

        for i in range(n):
            finishq.put(1)

        for thr in threads:
            thr.join()

        self.assertEqual(n, len(pool.resources))
        for resource in pool.resources:
            self.assertFalse(resource.claimed)
            self.assertEqual(1, len(resource.object))
            self.assertIn(resource.object[0], threads)

    def test_iteration(self):
        """
        Iteration over the pool resources, even when some are claimed,
        should eventually touch all resources (excluding ones created
        during iteration).
        """

        for i in range(25):
            started = Queue()
            n = 1000
            threads = []
            touched = []
            pool = EmptyListPool()
            rand = SystemRandom()

            def _run():
                psleep = rand.uniform(0.05, 0.1)
                with pool.transaction() as a:
                    started.put(1)
                    started.join()
                    a.append(rand.uniform(0, 1))
                    sleep(psleep)

            for i in range(n):
                th = Thread(target=_run)
                threads.append(th)
                th.start()

            for i in range(n):
                started.get()
                started.task_done()

            for resource in pool:
                touched.append(resource)

            for thr in threads:
                thr.join()

            self.assertItemsEqual(pool.resources, touched)

    def test_clear(self):
        """
        Clearing the pool should remove all resources known at the
        time of the call.
        """
        n = 10
        startq = Queue()
        finishq = Queue()
        rand = SystemRandom()
        threads = []
        pusher = None
        pool = SimplePool()

        def worker_run():
            with pool.transaction():
                startq.put(1)
                startq.join()
                sleep(rand.uniform(0, 0.5))
                finishq.get()
                finishq.task_done()

        def pusher_run():
            for i in range(n):
                finishq.put(1)
                sleep(rand.uniform(0, 0.1))
            finishq.join()

        # Allocate 10 resources in the pool by spinning up 10 threads
        for i in range(n):
            th = Thread(target=worker_run)
            threads.append(th)
            th.start()

        # Pull everything off the queue, allowing the workers to run
        for i in range(n):
            startq.get()
            startq.task_done()

        # Start the pusher that will allow them to proceed and exit
        pusher = Thread(target=pusher_run)
        threads.append(pusher)
        pusher.start()

        # Clear the pool
        pool.clear()

        # Wait for all threads to complete
        for t in threads:
            t.join()

        # Make sure that the pool resources are gone
        self.assertEqual(0, len(pool.resources))

    def test_stress(self):
        """
        Runs a large number of threads doing operations with resources
        checked out, ensuring properties of the pool.
        """
        rand = SystemRandom()
        n = rand.randint(1, 400)
        passes = rand.randint(1, 20)
        rounds = rand.randint(1, 200)
        breaker = rand.uniform(0, 1)
        pool = EmptyListPool()

        def _run():
            for i in range(rounds):
                with pool.transaction() as a:
                    self.assertEqual([], a)
                    a.append(currentThread())
                    self.assertEqual([currentThread()], a)

                    for p in range(passes):
                        self.assertEqual([currentThread()], a)
                        if rand.uniform(0, 1) > breaker:
                            break

                    a.remove(currentThread())

        threads = []

        for i in range(n):
            th = Thread(target=_run)
            threads.append(th)
            th.start()

        for th in threads:
            th.join()

if __name__ == '__main__':
    unittest.main()



from riak.client.transport import RiakClientTransport, \
    retryable, retryableHttpOnly
from riak.client.multiget import multiget
from riak.client.index_page import IndexPage
from riak.datatypes import TYPES
from riak.table import Table
from riak.util import bytes_to_str
from six import string_types, PY2


class RiakClientOperations(RiakClientTransport):
    """
    Methods for RiakClient that result in requests sent to the Riak
    cluster.

    Note that many of these methods have an implicit 'transport'
    argument that will be prepended automatically as part of the retry
    logic, and does not need to be supplied by the user.
    """

    @retryable
    def get_buckets(self, transport, bucket_type=None, timeout=None):
        """
        get_buckets(bucket_type=None, timeout=None)

        Get the list of buckets as :class:`RiakBucket
        <riak.bucket.RiakBucket>` instances.

        .. warning:: Do not use this in production, as it requires
           traversing through all keys stored in a cluster.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket_type: the optional containing bucket type
        :type bucket_type: :class:`~riak.bucket.BucketType`
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :rtype: list of :class:`RiakBucket <riak.bucket.RiakBucket>`
                instances
        """
        _validate_timeout(timeout)
        if bucket_type:
            bucketfn = self._bucket_type_bucket_builder
        else:
            bucketfn = self._default_type_bucket_builder

        return [bucketfn(bytes_to_str(name), bucket_type) for name in
                transport.get_buckets(bucket_type=bucket_type,
                                      timeout=timeout)]

    def stream_buckets(self, bucket_type=None, timeout=None):
        """
        Streams the list of buckets. This is a generator method that
        should be iterated over.

        .. warning:: Do not use this in production, as it requires
           traversing through all keys stored in a cluster.

        The caller should explicitly close the returned iterator,
        either using :func:`contextlib.closing` or calling ``close()``
        explicitly. Consuming the entire iterator will also close the
        stream. If it does not, the associated connection might not be
        returned to the pool. Example::

            from contextlib import closing

            # Using contextlib.closing
            with closing(client.stream_buckets()) as buckets:
                for bucket_list in buckets:
                    do_something(bucket_list)

            # Explicit close()
            stream = client.stream_buckets()
            for bucket_list in stream:
                 do_something(bucket_list)
            stream.close()

        :param bucket_type: the optional containing bucket type
        :type bucket_type: :class:`~riak.bucket.BucketType`
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :rtype: iterator that yields lists of :class:`RiakBucket
             <riak.bucket.RiakBucket>` instances

        """
        _validate_timeout(timeout)
        if bucket_type:
            bucketfn = self._bucket_type_bucket_builder
        else:
            bucketfn = self._default_type_bucket_builder

        resource = self._acquire()
        transport = resource.object
        stream = transport.stream_buckets(bucket_type=bucket_type,
                                          timeout=timeout)
        stream.attach(resource)
        try:
            for bucket_list in stream:
                bucket_list = [bucketfn(bytes_to_str(name), bucket_type)
                               for name in bucket_list]
                if len(bucket_list) > 0:
                    yield bucket_list
        finally:
            stream.close()

    @retryable
    def ping(self, transport):
        """
        ping()

        Check if the Riak server for this ``RiakClient`` instance is alive.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :rtype: boolean
        """
        return transport.ping()

    is_alive = ping

    @retryable
    def get_index(self, transport, bucket, index, startkey, endkey=None,
                  return_terms=None, max_results=None, continuation=None,
                  timeout=None, term_regex=None):
        """
        get_index(bucket, index, startkey, endkey=None, return_terms=None,\
                  max_results=None, continuation=None, timeout=None,\
                  term_regex=None)

        Queries a secondary index, returning matching keys.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket: the bucket whose index will be queried
        :type bucket: RiakBucket
        :param index: the index to query
        :type index: string
        :param startkey: the sole key to query, or beginning of the query range
        :type startkey: string, integer
        :param endkey: the end of the query range (optional if equality)
        :type endkey: string, integer
        :param return_terms: whether to include the secondary index value
        :type return_terms: boolean
        :param max_results: the maximum number of results to return (page size)
        :type max_results: integer
        :param continuation: the opaque continuation returned from a
            previous paginated request
        :type continuation: string
        :param timeout: a timeout value in milliseconds, or 'infinity'
        :type timeout: int
        :param term_regex: a regular expression used to filter index terms
        :type term_regex: string
        :rtype: :class:`~riak.client.index_page.IndexPage`
        """
        if timeout != 'infinity':
            _validate_timeout(timeout)

        page = IndexPage(self, bucket, index, startkey, endkey,
                         return_terms, max_results, term_regex)

        results, continuation = transport.get_index(
            bucket, index, startkey, endkey, return_terms=return_terms,
            max_results=max_results, continuation=continuation,
            timeout=timeout, term_regex=term_regex)

        page.results = results
        page.continuation = continuation
        return page

    def paginate_index(self, bucket, index, startkey, endkey=None,
                       max_results=1000, return_terms=None,
                       continuation=None, timeout=None, term_regex=None):
        """
        Iterates over a paginated index query. This is equivalent to calling
        :meth:`get_index` and then successively calling
        :meth:`~riak.client.index_page.IndexPage.next_page` until all
        results are exhausted.

        Because limiting the result set is necessary to invoke pagination,
        the ``max_results`` option has a default of ``1000``.

        :param bucket: the bucket whose index will be queried
        :type bucket: RiakBucket
        :param index: the index to query
        :type index: string
        :param startkey: the sole key to query, or beginning of the query range
        :type startkey: string, integer
        :param endkey: the end of the query range (optional if equality)
        :type endkey: string, integer
        :param return_terms: whether to include the secondary index value
        :type return_terms: boolean
        :param max_results: the maximum number of results to return (page
            size), defaults to 1000
        :type max_results: integer
        :param continuation: the opaque continuation returned from a
            previous paginated request
        :type continuation: string
        :param timeout: a timeout value in milliseconds, or 'infinity'
        :type timeout: int
        :param term_regex: a regular expression used to filter index terms
        :type term_regex: string
        :rtype: generator over instances of
          :class:`~riak.client.index_page.IndexPage`

        """
        page = self.get_index(bucket, index, startkey,
                              endkey=endkey, max_results=max_results,
                              return_terms=return_terms,
                              continuation=continuation,
                              timeout=timeout, term_regex=term_regex)
        yield page
        while page.has_next_page():
            page = page.next_page()
            yield page

    def stream_index(self, bucket, index, startkey, endkey=None,
                     return_terms=None, max_results=None, continuation=None,
                     timeout=None, term_regex=None):
        """
        Queries a secondary index, streaming matching keys through an
        iterator.

        The caller should explicitly close the returned iterator,
        either using :func:`contextlib.closing` or calling ``close()``
        explicitly. Consuming the entire iterator will also close the
        stream. If it does not, the associated connection might not be
        returned to the pool. Example::

            from contextlib import closing

            # Using contextlib.closing
            with closing(client.stream_index(mybucket, 'name_bin',
                                             'Smith')) as index:
                for key in index:
                    do_something(key)

            # Explicit close()
            stream = client.stream_index(mybucket, 'name_bin', 'Smith')
            for key in stream:
                 do_something(key)
            stream.close()

        :param bucket: the bucket whose index will be queried
        :type bucket: RiakBucket
        :param index: the index to query
        :type index: string
        :param startkey: the sole key to query, or beginning of the query range
        :type startkey: string, integer
        :param endkey: the end of the query range (optional if equality)
        :type endkey: string, integer
        :param return_terms: whether to include the secondary index value
        :type return_terms: boolean
        :param max_results: the maximum number of results to return (page size)
        :type max_results: integer
        :param continuation: the opaque continuation returned from a
            previous paginated request
        :type continuation: string
        :param timeout: a timeout value in milliseconds, or 'infinity'
        :type timeout: int
        :param term_regex: a regular expression used to filter index terms
        :type term_regex: string
        :rtype: :class:`~riak.client.index_page.IndexPage`

        """
        if timeout != 'infinity':
            _validate_timeout(timeout)

        page = IndexPage(self, bucket, index, startkey, endkey,
                         return_terms, max_results, term_regex)
        page.stream = True
        resource = self._acquire()
        transport = resource.object
        page.results = transport.stream_index(
            bucket, index, startkey, endkey, return_terms=return_terms,
            max_results=max_results, continuation=continuation,
            timeout=timeout, term_regex=term_regex)
        page.results.attach(resource)
        return page

    def paginate_stream_index(self, bucket, index, startkey, endkey=None,
                              max_results=1000, return_terms=None,
                              continuation=None, timeout=None,
                              term_regex=None):
        """
        Iterates over a streaming paginated index query. This is equivalent to
        calling :meth:`stream_index` and then successively calling
        :meth:`~riak.client.index_page.IndexPage.next_page` until all
        results are exhausted.

        Because limiting the result set is necessary to invoke
        pagination, the ``max_results`` option has a default of ``1000``.

        The caller should explicitly close each yielded page, either using
        :func:`contextlib.closing` or calling ``close()`` explicitly. Consuming
        the entire page will also close the stream. If it does not, the
        associated connection might not be returned to the pool. Example::

            from contextlib import closing

            # Using contextlib.closing
            for page in client.paginate_stream_index(mybucket, 'name_bin',
                                                     'Smith'):
                with closing(page):
                    for key in page:
                        do_something(key)

            # Explicit close()
            for page in client.paginate_stream_index(mybucket, 'name_bin',
                                                     'Smith'):
                for key in page:
                    do_something(key)
                page.close()

        :param bucket: the bucket whose index will be queried
        :type bucket: RiakBucket
        :param index: the index to query
        :type index: string
        :param startkey: the sole key to query, or beginning of the query range
        :type startkey: string, integer
        :param endkey: the end of the query range (optional if equality)
        :type endkey: string, integer
        :param return_terms: whether to include the secondary index value
        :type return_terms: boolean
        :param max_results: the maximum number of results to return (page
            size), defaults to 1000
        :type max_results: integer
        :param continuation: the opaque continuation returned from a
            previous paginated request
        :type continuation: string
        :param timeout: a timeout value in milliseconds, or 'infinity'
        :type timeout: int
        :param term_regex: a regular expression used to filter index terms
        :type term_regex: string
        :rtype: generator over instances of
          :class:`~riak.client.index_page.IndexPage`

        """
        page = self.stream_index(bucket, index, startkey,
                                 endkey=endkey,
                                 max_results=max_results,
                                 return_terms=return_terms,
                                 continuation=continuation,
                                 timeout=timeout,
                                 term_regex=term_regex)
        yield page
        while page.has_next_page():
            page = page.next_page()
            yield page

    @retryable
    def get_bucket_props(self, transport, bucket):
        """
        get_bucket_props(bucket)

        Fetches bucket properties for the given bucket.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket: the bucket whose properties will be fetched
        :type bucket: RiakBucket
        :rtype: dict
        """
        return transport.get_bucket_props(bucket)

    @retryable
    def set_bucket_props(self, transport, bucket, props):
        """
        set_bucket_props(bucket, props)

        Sets bucket properties for the given bucket.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket: the bucket whose properties will be set
        :type bucket: RiakBucket
        :param props: the properties to set
        :type props: dict
        """
        return transport.set_bucket_props(bucket, props)

    @retryable
    def clear_bucket_props(self, transport, bucket):
        """
        clear_bucket_props(bucket)

        Resets bucket properties for the given bucket.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket: the bucket whose properties will be set
        :type bucket: RiakBucket
        """
        return transport.clear_bucket_props(bucket)

    @retryable
    def get_bucket_type_props(self, transport, bucket_type):
        """
        get_bucket_type_props(bucket_type)

        Fetches properties for the given bucket-type.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket_type: the bucket-type whose properties will be fetched
        :type bucket_type: BucketType
        :rtype: dict
        """
        return transport.get_bucket_type_props(bucket_type)

    @retryable
    def set_bucket_type_props(self, transport, bucket_type, props):
        """
        set_bucket_type_props(bucket_type, props)

        Sets properties for the given bucket-type.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket_type: the bucket-type whose properties will be set
        :type bucket_type: BucketType
        :param props: the properties to set
        :type props: dict
        """
        return transport.set_bucket_type_props(bucket_type, props)

    @retryable
    def get_keys(self, transport, bucket, timeout=None):
        """
        get_keys(bucket, timeout=None)

        Lists all keys in a bucket.

        .. warning:: Do not use this in production, as it requires
           traversing through all keys stored in a cluster.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket: the bucket whose keys are fetched
        :type bucket: RiakBucket
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :rtype: list
        """
        _validate_timeout(timeout)
        return transport.get_keys(bucket, timeout=timeout)

    def stream_keys(self, bucket, timeout=None):
        """
        Lists all keys in a bucket via a stream. This is a generator
        method which should be iterated over.

        .. warning:: Do not use this in production, as it requires
           traversing through all keys stored in a cluster.

        The caller should explicitly close the returned iterator,
        either using :func:`contextlib.closing` or calling ``close()``
        explicitly. Consuming the entire iterator will also close the
        stream. If it does not, the associated connection might
        not be returned to the pool. Example::

            from contextlib import closing

            # Using contextlib.closing
            with closing(client.stream_keys(mybucket)) as keys:
                for key_list in keys:
                    do_something(key_list)

            # Explicit close()
            stream = client.stream_keys(mybucket)
            for key_list in stream:
                 do_something(key_list)
            stream.close()

        :param bucket: the bucket whose properties will be set
        :type bucket: RiakBucket
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :rtype: iterator
        """
        _validate_timeout(timeout)
        resource = self._acquire()
        transport = resource.object
        stream = transport.stream_keys(bucket, timeout=timeout)
        stream.attach(resource)
        try:
            for keylist in stream:
                if len(keylist) > 0:
                    if PY2:
                        yield keylist
                    else:
                        yield [bytes_to_str(item) for item in keylist]
        finally:
            stream.close()

    @retryable
    def put(self, transport, robj, w=None, dw=None, pw=None, return_body=None,
            if_none_match=None, timeout=None):
        """
        put(robj, w=None, dw=None, pw=None, return_body=None,\
            if_none_match=None, timeout=None)

        Stores an object in the Riak cluster.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param robj: the object to store
        :type robj: RiakObject
        :param w: the write quorum
        :type w: integer, string, None
        :param dw: the durable write quorum
        :type dw: integer, string, None
        :param pw: the primary write quorum
        :type pw: integer, string, None
        :param return_body: whether to return the resulting object
           after the write
        :type return_body: boolean
        :param if_none_match: whether to fail the write if the object
          exists
        :type if_none_match: boolean
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        """
        _validate_timeout(timeout)
        return transport.put(robj, w=w, dw=dw, pw=pw,
                             return_body=return_body,
                             if_none_match=if_none_match,
                             timeout=timeout)

    @retryable
    def ts_describe(self, transport, table):
        """
        ts_describe(table)

        Retrieve a time series table description from the Riak cluster.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param table: The timeseries table.
        :type table: string or :class:`Table <riak.table.Table>`
        :rtype: :class:`TsObject <riak.ts_object.TsObject>`
        """
        t = table
        if isinstance(t, string_types):
            t = Table(self, table)
        return transport.ts_describe(t)

    @retryable
    def ts_get(self, transport, table, key):
        """
        ts_get(table, key)

        Retrieve timeseries value by key

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param table: The timeseries table.
        :type table: string or :class:`Table <riak.table.Table>`
        :param key: The timeseries value's key.
        :type key: list
        :rtype: :class:`TsObject <riak.ts_object.TsObject>`
        """
        t = table
        if isinstance(t, string_types):
            t = Table(self, table)
        return transport.ts_get(t, key)

    @retryable
    def ts_put(self, transport, tsobj):
        """
        ts_put(tsobj)

        Stores time series data in the Riak cluster.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param tsobj: the time series object to store
        :type tsobj: RiakTsObject
        :rtype: boolean
        """
        return transport.ts_put(tsobj)

    @retryable
    def ts_delete(self, transport, table, key):
        """
        ts_delete(table, key)

        Delete timeseries value by key

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param table: The timeseries table.
        :type table: string or :class:`Table <riak.table.Table>`
        :param key: The timeseries value's key.
        :type key: list or dict
        :rtype: boolean
        """
        t = table
        if isinstance(t, string_types):
            t = Table(self, table)
        return transport.ts_delete(t, key)

    @retryable
    def ts_query(self, transport, table, query, interpolations=None):
        """
        ts_query(table, query, interpolations=None)

        Queries time series data in the Riak cluster.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param table: The timeseries table.
        :type table: string or :class:`Table <riak.table.Table>`
        :param query: The timeseries query.
        :type query: string
        :rtype: :class:`TsObject <riak.ts_object.TsObject>`
        """
        t = table
        if isinstance(t, string_types):
            t = Table(self, table)
        return transport.ts_query(t, query, interpolations)

    def ts_stream_keys(self, table, timeout=None):
        """
        Lists all keys in a time series table via a stream. This is a
        generator method which should be iterated over.

        The caller should explicitly close the returned iterator,
        either using :func:`contextlib.closing` or calling ``close()``
        explicitly. Consuming the entire iterator will also close the
        stream. If it does not, the associated connection might
        not be returned to the pool. Example::

            from contextlib import closing

            # Using contextlib.closing
            with closing(client.ts_stream_keys(mytable)) as keys:
                for key_list in keys:
                    do_something(key_list)

            # Explicit close()
            stream = client.ts_stream_keys(mytable)
            for key_list in stream:
                 do_something(key_list)
            stream.close()

        :param table: the table from which to stream keys
        :type table: Table
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :rtype: iterator
        """
        _validate_timeout(timeout)
        resource = self._acquire()
        transport = resource.object
        stream = transport.ts_stream_keys(table, timeout)
        stream.attach(resource)
        try:
            for keylist in stream:
                if len(keylist) > 0:
                    yield keylist
        finally:
            stream.close()

    @retryable
    def get(self, transport, robj, r=None, pr=None, timeout=None,
            basic_quorum=None, notfound_ok=None):
        """
        get(robj, r=None, pr=None, timeout=None)

        Fetches the contents of a Riak object.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param robj: the object to fetch
        :type robj: RiakObject
        :param r: the read quorum
        :type r: integer, string, None
        :param pr: the primary read quorum
        :type pr: integer, string, None
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :param basic_quorum: whether to use the "basic quorum" policy
           for not-founds
        :type basic_quorum: bool
        :param notfound_ok: whether to treat not-found responses as successful
        :type notfound_ok: bool
        """
        _validate_timeout(timeout)
        if not isinstance(robj.key, string_types):
            raise TypeError(
                'key must be a string, instead got {0}'.format(repr(robj.key)))

        return transport.get(robj, r=r, pr=pr, timeout=timeout,
                             basic_quorum=basic_quorum,
                             notfound_ok=notfound_ok)

    @retryable
    def delete(self, transport, robj, rw=None, r=None, w=None, dw=None,
               pr=None, pw=None, timeout=None):
        """
        delete(robj, rw=None, r=None, w=None, dw=None, pr=None, pw=None,\
               timeout=None)

        Deletes an object from Riak.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param robj: the object to delete
        :type robj: RiakObject
        :param rw: the read/write (delete) quorum
        :type rw: integer, string, None
        :param r: the read quorum
        :type r: integer, string, None
        :param pr: the primary read quorum
        :type pr: integer, string, None
        :param w: the write quorum
        :type w: integer, string, None
        :param dw: the durable write quorum
        :type dw: integer, string, None
        :param pw: the primary write quorum
        :type pw: integer, string, None
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        """
        _validate_timeout(timeout)
        return transport.delete(robj, rw=rw, r=r, w=w, dw=dw, pr=pr,
                                pw=pw, timeout=timeout)

    @retryable
    def mapred(self, transport, inputs, query, timeout):
        """
        mapred(inputs, query, timeout)

        Executes a MapReduce query.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param inputs: the input list/structure
        :type inputs: list, dict
        :param query: the list of query phases
        :type query: list
        :param timeout: the query timeout
        :type timeout: integer, None
        :rtype: mixed
        """
        _validate_timeout(timeout)
        return transport.mapred(inputs, query, timeout)

    def stream_mapred(self, inputs, query, timeout):
        """
        Streams a MapReduce query as (phase, data) pairs. This is a
        generator method which should be iterated over.

        The caller should explicitly close the returned iterator,
        either using :func:`contextlib.closing` or calling ``close()``
        explicitly. Consuming the entire iterator will also close the
        stream. If it does not, the associated connection might
        not be returned to the pool. Example::

            from contextlib import closing

            # Using contextlib.closing
            with closing(mymapred.stream()) as results:
                for phase, result in results:
                    do_something(phase, result)

            # Explicit close()
            stream = mymapred.stream()
            for phase, result in stream:
                 do_something(phase, result)
            stream.close()

        :param inputs: the input list/structure
        :type inputs: list, dict
        :param query: the list of query phases
        :type query: list
        :param timeout: the query timeout
        :type timeout: integer, None
        :rtype: iterator
        """
        _validate_timeout(timeout)
        resource = self._acquire()
        transport = resource.object
        stream = transport.stream_mapred(inputs, query, timeout)
        stream.attach(resource)
        try:
            for phase, data in stream:
                yield phase, data
        finally:
            stream.close()

    @retryable
    def create_search_index(self, transport, index, schema=None, n_val=None,
                            timeout=None):
        """
        create_search_index(index, schema=None, n_val=None)

        Create a search index of the given name, and optionally set
        a schema. If no schema is set, the default will be used.

        :param index: the name of the index to create
        :type index: string
        :param schema: the schema that this index will follow
        :type schema: string, None
        :param n_val: this indexes N value
        :type n_val: integer, None
        :param timeout: optional timeout (in ms)
        :type timeout: integer, None
        """
        return transport.create_search_index(index, schema, n_val, timeout)

    @retryable
    def get_search_index(self, transport, index):
        """
        get_search_index(index)

        Gets a search index of the given name if it exists, which will also
        return the schema. Raises a RiakError if no such schema exists. The
        returned dict contains keys ``'name'``, ``'schema'`` and
        ``'n_val'``.

        :param index: the name of the index to create
        :type index: string
        :rtype: dict
        """
        return transport.get_search_index(index)

    @retryable
    def list_search_indexes(self, transport):
        """list_search_indexes()

        Gets all search indexes and their schemas. The returned list
        contains dicts with keys ``'name'``, ``'schema'`` and ``'n_val'``.

        :return: list of dicts
        """
        return transport.list_search_indexes()

    @retryable
    def delete_search_index(self, transport, index):
        """
        delete_search_index(index)

        Delete the search index that matches the given name.

        :param index: the name of the index to delete
        :type index: string
        """
        return transport.delete_search_index(index)

    @retryable
    def create_search_schema(self, transport, schema, content):
        """
        create_search_schema(schema, content)

        Creates a Solr schema of the given name and content.
        Content must be valid Solr schema XML.

        :param schema: the name of the schema to create
        :type schema: string
        :param content: the solr schema xml content
        :type content: string
        """
        return transport.create_search_schema(schema, content)

    @retryable
    def get_search_schema(self, transport, schema):
        """
        get_search_schema(schema)

        Gets a search schema of the given name if it exists.
        Raises a RiakError if no such schema exists.  The schema is
        returned as a dict with keys ``'name'`` and ``'content'``.

        :param schema: the name of the schema to get
        :type schema: string

        :return: dict
        """
        return transport.get_search_schema(schema)

    @retryable
    def fulltext_search(self, transport, index, query, **params):
        """
        fulltext_search(index, query, **params)

        Performs a full-text search query.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param index: the bucket/index to search over
        :type index: string
        :param query: the search query
        :type query: string
        :param params: additional query flags
        :type params: dict
        :rtype: dict
        """
        return transport.search(index, query, **params)

    @retryableHttpOnly
    def fulltext_add(self, transport, index, docs):
        """
        fulltext_add(index, docs)

        .. deprecated:: 2.1.0 (Riak 2.0)
           Manual index maintenance is not supported for
           :ref:`Riak Search 2.0 <yz-label>`.

        Adds documents to the full-text index.

        .. note:: This request is automatically retried
           :attr:`retries` times if it fails due to network error.
           Only HTTP will be used for this request.

        :param index: the bucket/index in which to index these docs
        :type index: string
        :param docs: the list of documents
        :type docs: list
        """
        transport.fulltext_add(index, docs)

    @retryableHttpOnly
    def fulltext_delete(self, transport, index, docs=None, queries=None):
        """
        fulltext_delete(index, docs=None, queries=None)

        .. deprecated:: 2.1.0 (Riak 2.0)
           Manual index maintenance is not supported for
           :ref:`Riak Search 2.0 <yz-label>`.

        Removes documents from the full-text index.

        .. note:: This request is automatically retried
           :attr:`retries` times if it fails due to network error.
           Only HTTP will be used for this request.

        :param index: the bucket/index from which to delete
        :type index: string
        :param docs: a list of documents (with ids)
        :type docs: list
        :param queries: a list of queries to match and delete
        :type queries: list
        """
        transport.fulltext_delete(index, docs, queries)

    def multiget(self, pairs, **params):
        """Fetches many keys in parallel via threads.

        :param pairs: list of bucket_type/bucket/key tuple triples
        :type pairs: list
        :param params: additional request flags, e.g. r, pr
        :type params: dict
        :rtype: list of :class:`RiakObjects <riak.riak_object.RiakObject>`,
            :class:`Datatypes <riak.datatypes.Datatype>`, or tuples of
            bucket_type, bucket, key, and the exception raised on fetch
        """
        if self._multiget_pool:
            params['pool'] = self._multiget_pool
        return multiget(self, pairs, **params)

    @retryable
    def get_counter(self, transport, bucket, key, r=None, pr=None,
                    basic_quorum=None, notfound_ok=None):
        """get_counter(bucket, key, r=None, pr=None, basic_quorum=None,\
                       notfound_ok=None)

        Gets the value of a counter.

        .. deprecated:: 2.1.0 (Riak 2.0) Riak 1.4-style counters are
           deprecated in favor of the :class:`~riak.datatypes.Counter`
           datatype.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket: the bucket of the counter
        :type bucket: RiakBucket
        :param key: the key of the counter
        :type key: string
        :param r: the read quorum
        :type r: integer, string, None
        :param pr: the primary read quorum
        :type pr: integer, string, None
        :param basic_quorum: whether to use the "basic quorum" policy
           for not-founds
        :type basic_quorum: bool
        :param notfound_ok: whether to treat not-found responses as successful
        :type notfound_ok: bool
        :rtype: integer

        """
        return transport.get_counter(bucket, key, r=r, pr=pr)

    def update_counter(self, bucket, key, value, w=None, dw=None, pw=None,
                       returnvalue=False):
        """
        update_counter(bucket, key, value, w=None, dw=None, pw=None,\
                       returnvalue=False)

        .. deprecated:: 2.1.0 (Riak 2.0) Riak 1.4-style counters are
           deprecated in favor of the :class:`~riak.datatypes.Counter`
           datatype.

        Updates a counter by the given value. This operation is not
        idempotent and so should not be retried automatically.

        :param bucket: the bucket of the counter
        :type bucket: RiakBucket
        :param key: the key of the counter
        :type key: string
        :param value: the amount to increment or decrement
        :type value: integer
        :param w: the write quorum
        :type w: integer, string, None
        :param dw: the durable write quorum
        :type dw: integer, string, None
        :param pw: the primary write quorum
        :type pw: integer, string, None
        :param returnvalue: whether to return the updated value of the counter
        :type returnvalue: bool
        """
        if PY2:
            valid_types = (int, long)  # noqa
        else:
            valid_types = (int,)
        if type(value) not in valid_types:
            raise TypeError("Counter update amount must be an integer")
        if value == 0:
            raise ValueError("Cannot increment counter by 0")

        with self._transport() as transport:
            return transport.update_counter(bucket, key, value,
                                            w=w, dw=dw, pw=pw,
                                            returnvalue=returnvalue)

    increment_counter = update_counter

    def fetch_datatype(self, bucket, key, r=None, pr=None,
                       basic_quorum=None, notfound_ok=None,
                       timeout=None, include_context=None):
        """
        Fetches the value of a Riak Datatype.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket: the bucket of the datatype, which must belong to a
          :class:`~riak.bucket.BucketType`
        :type bucket: :class:`~riak.bucket.RiakBucket`
        :param key: the key of the datatype
        :type key: string
        :param r: the read quorum
        :type r: integer, string, None
        :param pr: the primary read quorum
        :type pr: integer, string, None
        :param basic_quorum: whether to use the "basic quorum" policy
           for not-founds
        :type basic_quorum: bool, None
        :param notfound_ok: whether to treat not-found responses as successful
        :type notfound_ok: bool, None
        :param timeout: a timeout value in milliseconds
        :type timeout: int, None
        :param include_context: whether to return the opaque context
          as well as the value, which is useful for removal operations
          on sets and maps
        :type include_context: bool, None
        :rtype: :class:`~riak.datatypes.Datatype`
        """
        dtype, value, context = self._fetch_datatype(
            bucket, key, r=r, pr=pr, basic_quorum=basic_quorum,
            notfound_ok=notfound_ok, timeout=timeout,
            include_context=include_context)

        return TYPES[dtype](bucket=bucket, key=key, value=value,
                            context=context)

    def update_datatype(self, datatype, w=None, dw=None, pw=None,
                        return_body=None, timeout=None,
                        include_context=None):
        """
        Sends an update to a Riak Datatype to the server. This operation is not
        idempotent and so will not be retried automatically.

        :param datatype: the datatype with pending updates
        :type datatype: :class:`~riak.datatypes.Datatype`
        :param w: the write quorum
        :type w: integer, string, None
        :param dw: the durable write quorum
        :type dw: integer, string, None
        :param pw: the primary write quorum
        :type pw: integer, string, None
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :param include_context: whether to return the opaque context
          as well as the value, which is useful for removal operations
          on sets and maps
        :type include_context: bool
        :rtype: tuple of datatype, opaque value and opaque context

        """
        _validate_timeout(timeout)

        with self._transport() as transport:
            return transport.update_datatype(datatype, w=w, dw=dw, pw=pw,
                                             return_body=return_body,
                                             timeout=timeout,
                                             include_context=include_context)

    @retryable
    def get_preflist(self, transport, bucket, key):
        """
        Fetch the preflist for a given bucket and key.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket: the bucket whose index will be queried
        :type bucket: RiakBucket
        :param key: the key of the preflist
        :type key: string

        :return: list of dicts (partition, node, primary)
        """
        return transport.get_preflist(bucket, key)

    def _bucket_type_bucket_builder(self, name, bucket_type):
        """
        Build a bucket from a bucket type

        :param name: Bucket name
        :param bucket_type: A bucket type
        :return: A bucket object
        """
        return bucket_type.bucket(name)

    def _default_type_bucket_builder(self, name, unused):
        """
        Build a bucket for the default bucket type

        :param name: Default bucket name
        :param unused: Unused
        :return: A bucket object
        """
        del unused  # Ignored parameters.
        return self.bucket(name)

    @retryable
    def _fetch_datatype(self, transport, bucket, key, r=None, pr=None,
                        basic_quorum=None, notfound_ok=None,
                        timeout=None, include_context=None):
        """
        _fetch_datatype(bucket, key, r=None, pr=None, basic_quorum=None,
                       notfound_ok=None, timeout=None, include_context=None)


        Fetches the value of a Riak Datatype as raw data. This is used
        internally to update already reified Datatype objects. Use the
        public version to fetch a reified type.

        .. note:: This request is automatically retried :attr:`retries`
           times if it fails due to network error.

        :param bucket: the bucket of the datatype, which must belong to a
          :class:`~riak.BucketType`
        :type bucket: RiakBucket
        :param key: the key of the datatype
        :type key: string, None
        :param r: the read quorum
        :type r: integer, string, None
        :param pr: the primary read quorum
        :type pr: integer, string, None
        :param basic_quorum: whether to use the "basic quorum" policy
           for not-founds
        :type basic_quorum: bool
        :param notfound_ok: whether to treat not-found responses as successful
        :type notfound_ok: bool
        :param timeout: a timeout value in milliseconds
        :type timeout: int
        :param include_context: whether to return the opaque context
          as well as the value, which is useful for removal operations
          on sets and maps
        :type include_context: bool
        :rtype: tuple of type, value and context
        """
        _validate_timeout(timeout)

        return transport.fetch_datatype(bucket, key, r=r, pr=pr,
                                        basic_quorum=basic_quorum,
                                        notfound_ok=notfound_ok,
                                        timeout=timeout,
                                        include_context=include_context)


def _validate_timeout(timeout):
    """
    Raises an exception if the given timeout is an invalid value.
    """
    if not (timeout is None or
            ((type(timeout) == int or
             (PY2 and type(timeout) == long)) and timeout > 0)):  # noqa
        raise ValueError("timeout must be a positive integer")



"""
Copyright 2015 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""


class RiakError(Exception):
    """
    Base class for exceptions generated in the Riak API.
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class ConflictError(RiakError):
    """
    Raised when an operation is attempted on a
    :class:`~riak.riak_object.RiakObject` that has more than one
    sibling.
    """
    def __init__(self, message="Object in conflict"):
        super(ConflictError, self).__init__(message)



from riak.test_server import TestServer
import unittest


class TestServerTestCase(unittest.TestCase):
    def setUp(self):
        self.test_server = TestServer()

    def tearDown(self):
        pass

    def test_options_defaults(self):
        self.assertEqual(
            self.test_server.app_config["riak_core"]["handoff_port"], 9001)
        self.assertEqual(
            self.test_server.app_config["riak_kv"]["pb_ip"], "127.0.0.1")

    def test_merge_riak_core_options(self):
        self.test_server = TestServer(riak_core={"handoff_port": 10000})
        self.assertEqual(
            self.test_server.app_config["riak_core"]["handoff_port"], 10000)

    def test_merge_riak_search_options(self):
        self.test_server = TestServer(
            riak_search={"search_backend": "riak_search_backend"})
        self.assertEqual(
            self.test_server.app_config["riak_search"]["search_backend"],
            "riak_search_backend")

    def test_merge_riak_kv_options(self):
        self.test_server = TestServer(riak_kv={"pb_ip": "192.168.2.1"})
        self.assertEqual(self.test_server.app_config["riak_kv"]["pb_ip"],
                         "192.168.2.1")

    def test_merge_vmargs(self):
        self.test_server = TestServer(vm_args={"-P": 65000})
        self.assertEqual(self.test_server.vm_args["-P"], 65000)

    def test_set_ring_state_dir(self):
        self.assertEqual(
            self.test_server.app_config["riak_core"]["ring_state_dir"],
            "/tmp/riak/test_server/data/ring")

    def test_set_default_tmp_dir(self):
        self.assertEqual(self.test_server.temp_dir, "/tmp/riak/test_server")

    def test_set_non_default_tmp_dir(self):
        tmp_dir = '/not/the/default/dir'
        server = TestServer(tmp_dir=tmp_dir)
        self.assertEqual(server.temp_dir, tmp_dir)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestServerTestCase())
    return suite



# -*- coding: utf-8 -*-
from __future__ import print_function
import platform
from riak.tests import RUN_SEARCH, RUN_YZ
from riak.tests.base import IntegrationTestBase

if platform.python_version() < '2.7':
    unittest = __import__('unittest2')
else:
    import unittest

testrun_search_bucket = 'searchbucket'


def setUpModule():
    if RUN_SEARCH and not RUN_YZ:
        c = IntegrationTestBase.create_client()
        b = c.bucket(testrun_search_bucket)
        b.enable_search()
        c.close()


def tearDownModule():
    if RUN_SEARCH and not RUN_YZ:
        c = IntegrationTestBase.create_client()
        b = c.bucket(testrun_search_bucket)
        b.clear_properties()
        c.close()


@unittest.skipUnless(RUN_SEARCH, 'RUN_SEARCH is 0')
class EnableSearchTests(IntegrationTestBase, unittest.TestCase):
    def test_bucket_search_enabled(self):
        bucket = self.client.bucket(self.bucket_name)
        self.assertFalse(bucket.search_enabled())

    def test_enable_search_commit_hook(self):
        bucket = self.client.bucket(testrun_search_bucket)
        bucket.clear_properties()

        c = self.create_client()
        self.assertFalse(c.bucket(testrun_search_bucket).search_enabled())
        c.close()

        bucket.enable_search()

        c = self.create_client()
        self.assertTrue(c.bucket(testrun_search_bucket).search_enabled())
        c.close()

    def test_disable_search_commit_hook(self):
        bucket = self.client.bucket(testrun_search_bucket)
        bucket.clear_properties()
        bucket.enable_search()

        c = self.create_client()
        self.assertTrue(c.bucket(testrun_search_bucket).search_enabled())
        c.close()

        bucket.disable_search()

        c = self.create_client()
        self.assertFalse(c.bucket(testrun_search_bucket).search_enabled())
        c.close()

        bucket.enable_search()


@unittest.skipUnless(RUN_SEARCH, 'RUN_SEARCH is 0')
class SolrSearchTests(IntegrationTestBase, unittest.TestCase):
    def test_add_document_to_index(self):
        self.client.fulltext_add(testrun_search_bucket,
                                 [{"id": "doc", "username": "tony"}])
        results = self.client.fulltext_search(testrun_search_bucket,
                                              "username:tony")
        self.assertEqual("tony", results['docs'][0]['username'])

    def test_add_multiple_documents_to_index(self):
        self.client.fulltext_add(
            testrun_search_bucket,
            [{"id": "dizzy", "username": "dizzy"},
             {"id": "russell", "username": "russell"}])
        results = self.client.fulltext_search(
            testrun_search_bucket, "username:russell OR username:dizzy")
        self.assertEqual(2, len(results['docs']))

    def test_delete_documents_from_search_by_id(self):
        self.client.fulltext_add(
            testrun_search_bucket,
            [{"id": "dizzy", "username": "dizzy"},
             {"id": "russell", "username": "russell"}])
        self.client.fulltext_delete(testrun_search_bucket, docs=["dizzy"])
        results = self.client.fulltext_search(
            testrun_search_bucket, "username:russell OR username:dizzy")
        self.assertEqual(1, len(results['docs']))

    def test_delete_documents_from_search_by_query(self):
        self.client.fulltext_add(
            testrun_search_bucket,
            [{"id": "dizzy", "username": "dizzy"},
             {"id": "russell", "username": "russell"}])
        self.client.fulltext_delete(
            testrun_search_bucket,
            queries=["username:dizzy", "username:russell"])
        results = self.client.fulltext_search(
            testrun_search_bucket, "username:russell OR username:dizzy")
        self.assertEqual(0, len(results['docs']))

    def test_delete_documents_from_search_by_query_and_id(self):
        self.client.fulltext_add(
            testrun_search_bucket,
            [{"id": "dizzy", "username": "dizzy"},
             {"id": "russell", "username": "russell"}])
        self.client.fulltext_delete(
            testrun_search_bucket,
            docs=["dizzy"],
            queries=["username:russell"])
        results = self.client.fulltext_search(
            testrun_search_bucket,
            "username:russell OR username:dizzy")
        self.assertEqual(0, len(results['docs']))


@unittest.skipUnless(RUN_SEARCH, 'RUN_SEARCH is 0')
class SearchTests(IntegrationTestBase, unittest.TestCase):
    def test_solr_search_from_bucket(self):
        bucket = self.client.bucket(testrun_search_bucket)
        bucket.new("user", {"username": "roidrage"}).store()
        results = bucket.search("username:roidrage")
        self.assertEqual(1, len(results['docs']))

    def test_solr_search_with_params_from_bucket(self):
        bucket = self.client.bucket(testrun_search_bucket)
        bucket.new("user", {"username": "roidrage"}).store()
        results = bucket.search("username:roidrage", wt="xml")
        self.assertEqual(1, len(results['docs']))

    def test_solr_search_with_params(self):
        bucket = self.client.bucket(testrun_search_bucket)
        bucket.new("user", {"username": "roidrage"}).store()
        results = self.client.fulltext_search(
            testrun_search_bucket,
            "username:roidrage", wt="xml")
        self.assertEqual(1, len(results['docs']))

    def test_solr_search(self):
        bucket = self.client.bucket(testrun_search_bucket)
        bucket.new("user", {"username": "roidrage"}).store()
        results = self.client.fulltext_search(testrun_search_bucket,
                                              "username:roidrage")
        self.assertEqual(1, len(results["docs"]))

    def test_search_integration(self):
        # Create some objects to search across...
        bucket = self.client.bucket(testrun_search_bucket)
        bucket.new("one", {"foo": "one", "bar": "red"}).store()
        bucket.new("two", {"foo": "two", "bar": "green"}).store()
        bucket.new("three", {"foo": "three", "bar": "blue"}).store()
        bucket.new("four", {"foo": "four", "bar": "orange"}).store()
        bucket.new("five", {"foo": "five", "bar": "yellow"}).store()

        # Run some operations...
        results = self.client.fulltext_search(testrun_search_bucket,
                                              "foo:one OR foo:two")
        if (len(results) == 0):
            print("\n\nNot running test \"testSearchIntegration()\".\n")
            print("""Please ensure that you have installed the Riak
            Search hook on bucket \"searchbucket\" by running
            \"bin/search-cmd install searchbucket\".\n\n""")
            return
        self.assertEqual(len(results['docs']), 2)
        query = "(foo:one OR foo:two OR foo:three OR foo:four) AND\
                 (NOT bar:green)"
        results = self.client.fulltext_search(testrun_search_bucket, query)

        self.assertEqual(len(results['docs']), 3)



from __future__ import print_function

import sys
import warnings

from collections import Mapping
from six import string_types, PY2


def is_timeseries_supported(v=None):
    if v is None:
        v = sys.version_info
    return v < (3,) or v >= (3, 4, 4)


def quacks_like_dict(object):
    """Check if object is dict-like"""
    return isinstance(object, Mapping)


def deep_merge(a, b):
    """Merge two deep dicts non-destructively

    Uses a stack to avoid maximum recursion depth exceptions

    >>> a = {'a': 1, 'b': {1: 1, 2: 2}, 'd': 6}
    >>> b = {'c': 3, 'b': {2: 7}, 'd': {'z': [1, 2, 3]}}
    >>> c = deep_merge(a, b)
    >>> from pprint import pprint; pprint(c)
    {'a': 1, 'b': {1: 1, 2: 7}, 'c': 3, 'd': {'z': [1, 2, 3]}}
    """
    assert quacks_like_dict(a), quacks_like_dict(b)
    dst = a.copy()

    stack = [(dst, b)]
    while stack:
        current_dst, current_src = stack.pop()
        for key in current_src:
            if key not in current_dst:
                current_dst[key] = current_src[key]
            else:
                if (quacks_like_dict(current_src[key]) and
                        quacks_like_dict(current_dst[key])):
                    stack.append((current_dst[key], current_src[key]))
                else:
                    current_dst[key] = current_src[key]
    return dst


def deprecated(message, stacklevel=3):
    """
    Prints a deprecation warning to the console.
    """
    warnings.warn(message, UserWarning, stacklevel=stacklevel)


class lazy_property(object):
    '''
    A method decorator meant to be used for lazy evaluation and
    memoization of an object attribute. The property should represent
    immutable data, as it replaces itself on first access.
    '''

    def __init__(self, fget):
        self.fget = fget
        self.func_name = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return None
        value = self.fget(obj)
        setattr(obj, self.func_name, value)
        return value


def decode_index_value(index, value):
    if "_int" in bytes_to_str(index):
        return str_to_long(value)
    elif PY2:
        return str(value)
    else:
        return bytes_to_str(value)


def bytes_to_str(value, encoding='utf-8'):
    if isinstance(value, string_types) or value is None:
        return value
    elif isinstance(value, list):
        return [bytes_to_str(elem) for elem in value]
    else:
        return value.decode(encoding)


def str_to_bytes(value, encoding='utf-8'):
    if PY2 or value is None:
        return value
    elif isinstance(value, list):
        return [str_to_bytes(elem) for elem in value]
    else:
        return value.encode(encoding)


def str_to_long(value, base=10):
    if value is None:
        return None
    elif PY2:
        return long(value, base)  # noqa
    else:
        return int(value, base)



# -*- coding: utf-8 -*-
import platform
from riak.riak_object import RiakObject
from riak.bucket import RiakBucket, BucketType
from riak.tests.base import IntegrationTestBase

if platform.python_version() < '2.7':
    unittest = __import__('unittest2')
else:
    import unittest


class BucketTypeRichComparisonTest(unittest.TestCase):
    def test_btype_eq(self):
        a = BucketType('client', 'a')
        b = BucketType('client', 'a')
        c = BucketType(None, 'a')
        d = BucketType(None, 'a')
        self.assertEqual(a, b)
        self.assertEqual(c, d)

    def test_btype_nq(self):
        a = BucketType('client', 'a')
        b = BucketType('client', 'b')
        c = BucketType(None, 'a')
        d = BucketType(None, 'a')
        self.assertNotEqual(a, b, "matched with different name, same client")
        self.assertNotEqual(a, c, "matched with different client, same name")
        self.assertNotEqual(b, d, "matched with nothing in common")

    def test_btype_hash(self):
        a = BucketType('client', 'a')
        b = BucketType('client', 'a')
        c = BucketType('client', 'c')
        d = BucketType('client2', 'a')
        self.assertEqual(hash(a), hash(b),
                         'same bucket type has different hashes')
        self.assertNotEqual(hash(a), hash(c),
                            'different bucket has same hash')
        self.assertNotEqual(hash(a), hash(d),
                            'same bucket type, different client has same hash')


class RiakBucketRichComparisonTest(unittest.TestCase):
    def test_bucket_eq(self):
        default_bt = BucketType(None, "default")
        foo_bt = BucketType(None, "foo")
        a = RiakBucket('client', 'a', default_bt)
        b = RiakBucket('client', 'a', default_bt)
        c = RiakBucket('client', 'a', foo_bt)
        d = RiakBucket('client', 'a', foo_bt)
        self.assertEqual(a, b)
        self.assertEqual(c, d)

    def test_bucket_nq(self):
        default_bt = BucketType(None, "default")
        foo_bt = BucketType(None, "foo")
        a = RiakBucket('client', 'a', default_bt)
        b = RiakBucket('client', 'b', default_bt)
        c = RiakBucket('client', 'a', foo_bt)
        self.assertNotEqual(a, b, 'matched with a different bucket')
        self.assertNotEqual(a, c, 'matched with a different bucket type')

    def test_bucket_hash(self):
        default_bt = BucketType(None, "default")
        foo_bt = BucketType(None, "foo")
        a = RiakBucket('client', 'a', default_bt)
        b = RiakBucket('client', 'a', default_bt)
        c = RiakBucket('client', 'c', default_bt)
        d = RiakBucket('client', 'a', foo_bt)
        self.assertEqual(hash(a), hash(b),
                         'same bucket has different hashes')
        self.assertNotEqual(hash(a), hash(c),
                            'different bucket has same hash')
        self.assertNotEqual(hash(a), hash(d),
                            'same bucket, different bucket type has same hash')


class RiakObjectComparisonTest(unittest.TestCase):
    def test_object_eq(self):
        a = RiakObject(None, 'bucket', 'key')
        b = RiakObject(None, 'bucket', 'key')
        self.assertEqual(a, b)
        default_bt = BucketType(None, "default")
        bucket_a = RiakBucket('client', 'a', default_bt)
        bucket_b = RiakBucket('client', 'a', default_bt)
        c = RiakObject(None, bucket_a, 'key')
        d = RiakObject(None, bucket_b, 'key')
        self.assertEqual(c, d)

    def test_object_nq(self):
        a = RiakObject(None, 'bucket', 'key')
        b = RiakObject(None, 'bucket', 'not key')
        c = RiakObject(None, 'not bucket', 'key')
        self.assertNotEqual(a, b, 'matched with different keys')
        self.assertNotEqual(a, c, 'matched with different buckets')
        default_bt = BucketType(None, "default")
        foo_bt = BucketType(None, "foo")
        bucket_a = RiakBucket('client', 'a', default_bt)
        bucket_b = RiakBucket('client', 'a', foo_bt)
        c = RiakObject(None, bucket_a, 'key')
        d = RiakObject(None, bucket_b, 'key')
        self.assertNotEqual(c, d)

    def test_object_hash(self):
        a = RiakObject(None, 'bucket', 'key')
        b = RiakObject(None, 'bucket', 'key')
        c = RiakObject(None, 'bucket', 'not key')
        self.assertEqual(hash(a), hash(b), 'same object has different hashes')
        self.assertNotEqual(hash(a), hash(c), 'different object has same hash')

        default_bt = BucketType(None, "default")
        foo_bt = BucketType(None, "foo")
        bucket_a = RiakBucket('client', 'a', default_bt)
        bucket_b = RiakBucket('client', 'a', foo_bt)
        d = RiakObject(None, bucket_a, 'key')
        e = RiakObject(None, bucket_a, 'key')
        f = RiakObject(None, bucket_b, 'key')
        g = RiakObject(None, bucket_b, 'not key')
        self.assertEqual(hash(d), hash(e),
                         'same object, same bucket_type has different hashes')
        self.assertNotEqual(hash(e), hash(f),
                            'same object, different bucket type has the '
                            'same hash')
        self.assertNotEqual(hash(d), hash(g),
                            'different object, different bucket '
                            'type has same hash')

    def test_object_valid_key(self):
        a = RiakObject(None, 'bucket', 'key')
        self.assertIsInstance(a, RiakObject, 'valid key name is rejected')
        try:
            b = RiakObject(None, 'bucket', '')
        except ValueError:
            b = None
        self.assertIsNone(b, 'empty object key not allowed')


class RiakClientComparisonTest(IntegrationTestBase, unittest.TestCase):
    def test_client_eq(self):
        self.protocol = 'http'
        a = self.create_client(host='host1', http_port=11)
        b = self.create_client(host='host1', http_port=11)
        self.assertEqual(a, b)
        a.close()
        b.close()

    def test_client_nq(self):
        self.protocol = 'http'
        a = self.create_client(host='host1', http_port=11)
        b = self.create_client(host='host2', http_port=11)
        c = self.create_client(host='host1', http_port=12)
        self.assertNotEqual(a, b, 'matched with different hosts')
        self.assertNotEqual(a, c, 'matched with different ports')
        a.close()
        b.close()
        c.close()

    def test_client_hash(self):
        self.protocol = 'http'
        a = self.create_client(host='host1', http_port=11)
        b = self.create_client(host='host1', http_port=11)
        c = self.create_client(host='host2', http_port=11)
        self.assertEqual(hash(a), hash(b), 'same object has different hashes')
        self.assertNotEqual(hash(a), hash(c), 'different object has same hash')
        a.close()
        b.close()
        c.close()

if __name__ == '__main__':
    unittest.main()



import platform
from riak import RiakError, RiakObject
from riak.bucket import RiakBucket, BucketType
from riak.tests import RUN_BTYPES
from riak.tests.base import IntegrationTestBase
from riak.tests.comparison import Comparison

if platform.python_version() < '2.7':
    unittest = __import__('unittest2')
else:
    import unittest


@unittest.skipUnless(RUN_BTYPES, "RUN_BTYPES is 0")
class BucketTypeTests(IntegrationTestBase, unittest.TestCase, Comparison):
    def test_btype_init(self):
        btype = self.client.bucket_type('foo')
        self.assertIsInstance(btype, BucketType)
        self.assertEqual('foo', btype.name)
        self.assertIs(btype, self.client.bucket_type('foo'))

    def test_btype_get_bucket(self):
        btype = self.client.bucket_type('foo')
        bucket = btype.bucket(self.bucket_name)
        self.assertIsInstance(bucket, RiakBucket)
        self.assertIs(btype, bucket.bucket_type)
        self.assertIs(bucket,
                      self.client.bucket_type('foo').bucket(self.bucket_name))
        self.assertIsNot(bucket, self.client.bucket(self.bucket_name))

    def test_btype_default(self):
        defbtype = self.client.bucket_type('default')
        othertype = self.client.bucket_type('foo')
        self.assertTrue(defbtype.is_default())
        self.assertFalse(othertype.is_default())

    def test_btype_repr(self):
        defbtype = self.client.bucket_type("default")
        othertype = self.client.bucket_type("foo")
        self.assertEqual("<BucketType 'default'>", str(defbtype))
        self.assertEqual("<BucketType 'foo'>", str(othertype))
        self.assertEqual("<BucketType 'default'>", repr(defbtype))
        self.assertEqual("<BucketType 'foo'>", repr(othertype))

    def test_btype_get_props(self):
        defbtype = self.client.bucket_type("default")
        btype = self.client.bucket_type('no_siblings')
        with self.assertRaises(ValueError):
            defbtype.get_properties()

        props = btype.get_properties()
        self.assertIsInstance(props, dict)
        self.assertIn('n_val', props)
        self.assertEqual(3, props['n_val'])

    def test_btype_set_props(self):
        defbtype = self.client.bucket_type("default")
        btype = self.client.bucket_type('no_siblings')
        with self.assertRaises(ValueError):
            defbtype.set_properties({'allow_mult': True})

        oldprops = btype.get_properties()
        try:
            btype.set_properties({'allow_mult': True})
            newprops = btype.get_properties()
            self.assertIsInstance(newprops, dict)
            self.assertIn('allow_mult', newprops)
            self.assertTrue(newprops['allow_mult'])
            if 'claimant' in oldprops:  # HTTP hack
                del oldprops['claimant']
        finally:
            btype.set_properties(oldprops)

    def test_btype_set_props_immutable(self):
        btype = self.client.bucket_type("maps")
        with self.assertRaises(RiakError):
            btype.set_property('datatype', 'counter')

    def test_btype_list_buckets(self):
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket(self.bucket_name)
        obj = bucket.new(self.key_name)
        obj.data = [1, 2, 3]
        obj.store()

        self.assertIn(bucket, btype.get_buckets())
        buckets = []
        for nested_buckets in btype.stream_buckets():
            buckets.extend(nested_buckets)

        self.assertIn(bucket, buckets)

    def test_btype_list_keys(self):
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket(self.bucket_name)

        obj = bucket.new(self.key_name)
        obj.data = [1, 2, 3]
        obj.store()

        self.assertIn(self.key_name, bucket.get_keys())
        keys = []
        for keylist in bucket.stream_keys():
            keys.extend(keylist)

        self.assertIn(self.key_name, keys)

    def test_default_btype_list_buckets(self):
        default_btype = self.client.bucket_type("default")
        bucket = default_btype.bucket(self.bucket_name)
        obj = bucket.new(self.key_name)
        obj.data = [1, 2, 3]
        obj.store()

        self.assertIn(bucket, default_btype.get_buckets())
        buckets = []
        for nested_buckets in default_btype.stream_buckets():
            buckets.extend(nested_buckets)

        self.assertIn(bucket, buckets)

        self.assertItemsEqual(buckets, self.client.get_buckets())

    def test_default_btype_list_keys(self):
        btype = self.client.bucket_type("default")
        bucket = btype.bucket(self.bucket_name)

        obj = bucket.new(self.key_name)
        obj.data = [1, 2, 3]
        obj.store()

        self.assertIn(self.key_name, bucket.get_keys())
        keys = []
        for keylist in bucket.stream_keys():
            keys.extend(keylist)

        self.assertIn(self.key_name, keys)

        oldapikeys = self.client.get_keys(self.client.bucket(self.bucket_name))
        self.assertItemsEqual(keys, oldapikeys)

    def test_multiget_bucket_types(self):
        btype = self.client.bucket_type('no_siblings')
        bucket = btype.bucket(self.bucket_name)

        for i in range(100):
            obj = bucket.new(self.key_name + str(i))
            obj.data = {'id': i}
            obj.store()

        mget = bucket.multiget([self.key_name + str(i) for i in range(100)])
        for mobj in mget:
            self.assertIsInstance(mobj, RiakObject)
            self.assertEqual(bucket, mobj.bucket)
            self.assertEqual(btype, mobj.bucket.bucket_type)

    def test_write_once_bucket_type(self):
        btype = self.client.bucket_type('write_once')
        bucket = btype.bucket(self.bucket_name)

        for i in range(100):
            obj = bucket.new(self.key_name + str(i))
            obj.data = {'id': i}
            obj.store()

        mget = bucket.multiget([self.key_name + str(i) for i in range(100)])
        for mobj in mget:
            self.assertIsInstance(mobj, RiakObject)
            self.assertEqual(bucket, mobj.bucket)
            self.assertEqual(btype, mobj.bucket.bucket_type)

        props = btype.get_properties()
        self.assertIn('write_once', props)
        self.assertEqual(True, props['write_once'])



"""
Copyright 2015 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import collections
from .datatype import Datatype
from six import string_types
from riak.datatypes import TYPES

__all__ = ['Set']


class Set(collections.Set, Datatype):
    """A convergent datatype representing a Set with observed-remove
    semantics. Currently strings are the only supported value type.
    Example::

        myset.add('barista')
        myset.add('roaster')
        myset.add('brewer')

    Likewise they can simply be removed::

        myset.discard('barista')

    This datatype also implements the `Set ABC
    <https://docs.python.org/2/library/collections.html>`_, meaning it
    supports ``len()``, ``in``, and iteration.

    """

    type_name = 'set'
    _type_error_msg = "Sets can only be iterables of strings"

    def _post_init(self):
        self._adds = set()
        self._removes = set()

    def _default_value(self):
        return frozenset()

    @Datatype.modified.getter
    def modified(self):
        """
        Whether this set has staged adds or removes.
        """
        return len(self._removes | self._adds) > 0

    def to_op(self):
        """
        Extracts the modification operation from the set.

        :rtype: dict, None
        """
        if not self._adds and not self._removes:
            return None
        changes = {}
        if self._adds:
            changes['adds'] = list(self._adds)
        if self._removes:
            changes['removes'] = list(self._removes)
        return changes

    # collections.Set API, operates only on the immutable version
    def __contains__(self, element):
        return element in self.value

    def __iter__(self):
        return iter(self.value)

    def __len__(self):
        return len(self.value)

    # Sort of like collections.MutableSet API, without the additional
    # methods.
    def add(self, element):
        """
        Adds an element to the set.

        .. note: You may add elements that already exist in the set.
           This may be used as an "assertion" that the element is a
           member.

        :param element: the element to add
        :type element: str
        """
        _check_element(element)
        self._adds.add(element)

    def discard(self, element):
        """
        Removes an element from the set.

        .. note: You may remove elements from the set that are not
           present, but a context from the server is required.

        :param element: the element to remove
        :type element: str
        """
        _check_element(element)
        self._require_context()
        self._removes.add(element)

    def _coerce_value(self, new_value):
        return frozenset(new_value)

    def _check_type(self, new_value):
        if not isinstance(new_value, collections.Iterable):
            return False
        for element in new_value:
            if not isinstance(element, string_types):
                return False
        return True


def _check_element(element):
    if not isinstance(element, string_types):
        raise TypeError("Set elements can only be strings")


TYPES['set'] = Set



# -*- coding: utf-8 -*-
import os
import platform
from six import string_types, PY2, PY3

import copy
from time import sleep
from riak import ConflictError, RiakBucket, RiakError
from riak.resolver import default_resolver, last_written_resolver
from riak.tests import RUN_RESOLVE
from riak.tests.base import IntegrationTestBase
from riak.tests.comparison import Comparison

try:
    import simplejson as json
except ImportError:
    import json

if platform.python_version() < '2.7':
    unittest = __import__('unittest2')
else:
    import unittest

if PY2:
    import cPickle
    test_pickle_dumps = cPickle.dumps
    test_pickle_loads = cPickle.loads
else:
    import pickle
    test_pickle_dumps = pickle.dumps
    test_pickle_loads = pickle.loads


testrun_sibs_bucket = 'sibsbucket'
testrun_props_bucket = 'propsbucket'


def setUpModule():
    c = IntegrationTestBase.create_client()
    c.bucket(testrun_sibs_bucket).allow_mult = True
    c.close()


def tearDownModule():
    c = IntegrationTestBase.create_client()
    c.bucket(testrun_sibs_bucket).clear_properties()
    c.bucket(testrun_props_bucket).clear_properties()
    c.close()


class NotJsonSerializable(object):

    def __init__(self, *args, **kwargs):
        self.args = list(args)
        self.kwargs = kwargs

    def __eq__(self, other):
        if len(self.args) != len(other.args):
            return False
        if len(self.kwargs) != len(other.kwargs):
            return False
        for name, value in self.kwargs.items():
            if other.kwargs[name] != value:
                return False
        value1_args = copy.copy(self.args)
        value2_args = copy.copy(other.args)
        value1_args.sort()
        value2_args.sort()
        for i in range(len(value1_args)):
            if value1_args[i] != value2_args[i]:
                return False
        return True


class BasicKVTests(IntegrationTestBase, unittest.TestCase, Comparison):
    def test_no_returnbody(self):
        bucket = self.client.bucket(self.bucket_name)
        o = bucket.new(self.key_name, "bar").store(return_body=False)
        self.assertEqual(o.vclock, None)

    def test_many_link_headers_should_work_fine(self):
        bucket = self.client.bucket(self.bucket_name)
        o = bucket.new("lots_of_links", "My god, it's full of links!")
        for i in range(0, 300):
            link = ("other", "key%d" % i, "next")
            o.add_link(link)

        o.store()
        stored_object = bucket.get("lots_of_links")
        self.assertEqual(len(stored_object.links), 300)

    def test_is_alive(self):
        self.assertTrue(self.client.is_alive())

    def test_store_and_get(self):
        bucket = self.client.bucket(self.bucket_name)
        rand = self.randint()
        obj = bucket.new('foo', rand)
        obj.store()
        obj = bucket.get('foo')
        self.assertTrue(obj.exists)
        self.assertEqual(obj.bucket.name, self.bucket_name)
        self.assertEqual(obj.key, 'foo')
        self.assertEqual(obj.data, rand)

        # unicode objects are fine, as long as they don't
        # contain any non-ASCII chars
        if PY2:
            self.client.bucket(unicode(self.bucket_name))  # noqa
        else:
            self.client.bucket(self.bucket_name)
        if PY2:
            self.assertRaises(TypeError, self.client.bucket, u'búcket')
            self.assertRaises(TypeError, self.client.bucket, 'búcket')
        else:
            self.client.bucket(u'búcket')
            self.client.bucket('búcket')

        bucket.get(u'foo')
        if PY2:
            self.assertRaises(TypeError, bucket.get, u'føø')
            self.assertRaises(TypeError, bucket.get, 'føø')

            self.assertRaises(TypeError, bucket.new, u'foo', 'éå')
            self.assertRaises(TypeError, bucket.new, u'foo', 'éå')
            self.assertRaises(TypeError, bucket.new, 'foo', u'éå')
            self.assertRaises(TypeError, bucket.new, 'foo', u'éå')
        else:
            bucket.get(u'føø')
            bucket.get('føø')

            bucket.new(u'foo', 'éå')
            bucket.new(u'foo', 'éå')
            bucket.new('foo', u'éå')
            bucket.new('foo', u'éå')

        obj2 = bucket.new('baz', rand, 'application/json')
        obj2.charset = 'UTF-8'
        obj2.store()
        obj2 = bucket.get('baz')
        self.assertEqual(obj2.data, rand)

    def test_store_obj_with_unicode(self):
        bucket = self.client.bucket(self.bucket_name)
        data = {u'føø': u'éå'}
        obj = bucket.new('foo', data)
        obj.store()
        obj = bucket.get('foo')
        self.assertEqual(obj.data, data)

    def test_store_unicode_string(self):
        bucket = self.client.bucket(self.bucket_name)
        data = u"some unicode data: \u00c6"
        obj = bucket.new(self.key_name, encoded_data=data.encode('utf-8'),
                         content_type='text/plain')
        obj.charset = 'utf-8'
        obj.store()
        obj2 = bucket.get(self.key_name)
        self.assertEqual(data, obj2.encoded_data.decode('utf-8'))

    def test_string_bucket_name(self):
        # Things that are not strings cannot be bucket names
        for bad in (12345, True, None, {}, []):
            with self.assert_raises_regex(TypeError, 'must be a string'):
                self.client.bucket(bad)

            with self.assert_raises_regex(TypeError, 'must be a string'):
                RiakBucket(self.client, bad, None)

        # Unicode bucket names are not supported in Python 2.x,
        # if they can't be encoded to ASCII. This should be changed in a
        # future  release.
        if PY2:
            with self.assert_raises_regex(TypeError,
                                          'Unicode bucket names '
                                          'are not supported'):
                self.client.bucket(u'føø')
        else:
            self.client.bucket(u'føø')

        # This is fine, since it's already ASCII
        self.client.bucket('ASCII')

    def test_generate_key(self):
        # Ensure that Riak generates a random key when
        # the key passed to bucket.new() is None.
        bucket = self.client.bucket('random_key_bucket')
        existing_keys = bucket.get_keys()
        o = bucket.new(None, data={})
        self.assertIsNone(o.key)
        o.store()
        self.assertIsNotNone(o.key)
        self.assertNotIn('/', o.key)
        self.assertNotIn(o.key, existing_keys)
        self.assertEqual(len(bucket.get_keys()), len(existing_keys) + 1)

    def test_stream_keys(self):
        bucket = self.client.bucket('random_key_bucket')
        regular_keys = bucket.get_keys()
        self.assertNotEqual(len(regular_keys), 0)
        streamed_keys = []
        for keylist in bucket.stream_keys():
            self.assertNotEqual([], keylist)
            for key in keylist:
                self.assertIsInstance(key, string_types)
            streamed_keys += keylist
        self.assertEqual(sorted(regular_keys), sorted(streamed_keys))

    def test_stream_keys_timeout(self):
        bucket = self.client.bucket('random_key_bucket')
        for key in range(1, 1000):
            o = bucket.new(None, data={})
            o.store()
        streamed_keys = []
        with self.assertRaises(RiakError):
            for keylist in self.client.stream_keys(bucket, timeout=1):
                self.assertNotEqual([], keylist)
                for key in keylist:
                    self.assertIsInstance(key, string_types)
                streamed_keys += keylist

    def test_stream_keys_abort(self):
        bucket = self.client.bucket('random_key_bucket')
        regular_keys = bucket.get_keys()
        self.assertNotEqual(len(regular_keys), 0)
        try:
            for keylist in bucket.stream_keys():
                raise RuntimeError("abort")
        except RuntimeError:
            pass

        # If the stream was closed correctly, this will not error
        robj = bucket.get(regular_keys[0])
        self.assertEqual(len(robj.siblings), 1)
        self.assertEqual(True, robj.exists)

    def test_bad_key(self):
        bucket = self.client.bucket(self.bucket_name)
        obj = bucket.new()
        with self.assertRaises(TypeError):
            bucket.get(None)

        with self.assertRaises(TypeError):
            self.client.get(obj)

        with self.assertRaises(TypeError):
            bucket.get(1)

    def test_binary_store_and_get(self):
        bucket = self.client.bucket(self.bucket_name)
        # Store as binary, retrieve as binary, then compare...
        rand = str(self.randint())
        if PY2:
            rand = bytes(rand)
        else:
            rand = bytes(rand, 'utf-8')
        obj = bucket.new(self.key_name, encoded_data=rand,
                         content_type='text/plain')
        obj.store()
        obj = bucket.get(self.key_name)
        self.assertTrue(obj.exists)
        self.assertEqual(obj.encoded_data, rand)
        # Store as JSON, retrieve as binary, JSON-decode, then compare...
        data = [self.randint(), self.randint(), self.randint()]
        key2 = self.randname()
        obj = bucket.new(key2, data)
        obj.store()
        obj = bucket.get(key2)
        self.assertEqual(data, json.loads(obj.encoded_data.decode()))

    def test_blank_binary_204(self):
        bucket = self.client.bucket(self.bucket_name)

        # this should *not* raise an error
        empty = ""
        if PY2:
            empty = bytes(empty)
        else:
            empty = bytes(empty, 'utf-8')
        obj = bucket.new('foo2', encoded_data=empty, content_type='text/plain')
        obj.store()
        obj = bucket.get('foo2')
        self.assertTrue(obj.exists)
        self.assertEqual(obj.encoded_data, empty)

    def test_custom_bucket_encoder_decoder(self):
        bucket = self.client.bucket(self.bucket_name)
        # Teach the bucket how to pickle
        bucket.set_encoder('application/x-pickle', test_pickle_dumps)
        bucket.set_decoder('application/x-pickle', test_pickle_loads)
        data = {'array': [1, 2, 3], 'badforjson': NotJsonSerializable(1, 3)}
        obj = bucket.new(self.key_name, data, 'application/x-pickle')
        obj.store()
        obj2 = bucket.get(self.key_name)
        self.assertEqual(data, obj2.data)

    def test_custom_client_encoder_decoder(self):
        bucket = self.client.bucket(self.bucket_name)
        # Teach the client how to pickle
        self.client.set_encoder('application/x-pickle', test_pickle_dumps)
        self.client.set_decoder('application/x-pickle', test_pickle_loads)
        data = {'array': [1, 2, 3], 'badforjson': NotJsonSerializable(1, 3)}
        obj = bucket.new(self.key_name, data, 'application/x-pickle')
        obj.store()
        obj2 = bucket.get(self.key_name)
        self.assertEqual(data, obj2.data)

    def test_unknown_content_type_encoder_decoder(self):
        # Bypass the content_type encoders
        bucket = self.client.bucket(self.bucket_name)
        data = "some funny data"
        if PY3:
            # Python 3.x needs to store binaries
            data = data.encode()
        obj = bucket.new(self.key_name,
                         encoded_data=data,
                         content_type='application/x-frobnicator')
        obj.store()
        obj2 = bucket.get(self.key_name)
        self.assertEqual(data, obj2.encoded_data)

    def test_text_plain_encoder_decoder(self):
        bucket = self.client.bucket(self.bucket_name)
        data = "some funny data"
        obj = bucket.new(self.key_name, data, content_type='text/plain')
        obj.store()
        obj2 = bucket.get(self.key_name)
        self.assertEqual(data, obj2.data)

    def test_missing_object(self):
        bucket = self.client.bucket(self.bucket_name)
        obj = bucket.get(self.key_name)
        self.assertFalse(obj.exists)
        # Object with no siblings should not raise the ConflictError
        self.assertIsNone(obj.data)

    def test_delete(self):
        bucket = self.client.bucket(self.bucket_name)
        rand = self.randint()
        obj = bucket.new(self.key_name, rand)
        obj.store()
        obj = bucket.get(self.key_name)
        self.assertTrue(obj.exists)

        obj.delete()
        obj.reload()
        self.assertFalse(obj.exists)

    def test_bucket_delete(self):
        bucket = self.client.bucket(self.bucket_name)
        rand = self.randint()
        obj = bucket.new(self.key_name, rand)
        obj.store()

        bucket.delete(self.key_name)
        obj.reload()
        self.assertFalse(obj.exists)

    def test_set_bucket_properties(self):
        bucket = self.client.bucket(testrun_props_bucket)
        # Test setting allow mult...
        bucket.allow_mult = True
        # Test setting nval...
        bucket.n_val = 1

        c2 = self.create_client()
        bucket2 = c2.bucket(testrun_props_bucket)
        self.assertTrue(bucket2.allow_mult)
        self.assertEqual(bucket2.n_val, 1)
        # Test setting multiple properties...
        bucket.set_properties({"allow_mult": False, "n_val": 2})

        c3 = self.create_client()
        bucket3 = c3.bucket(testrun_props_bucket)
        self.assertFalse(bucket3.allow_mult)
        self.assertEqual(bucket3.n_val, 2)

        # clean up!
        c2.close()
        c3.close()

    def test_if_none_match(self):
        bucket = self.client.bucket(self.bucket_name)
        obj = bucket.get(self.key_name)
        obj.delete()

        obj.reload()
        self.assertFalse(obj.exists)
        obj.data = ["first store"]
        obj.content_type = 'application/json'
        obj.store()

        obj.data = ["second store"]
        with self.assertRaises(Exception):
            obj.store(if_none_match=True)

    def test_siblings(self):
        # Set up the bucket, clear any existing object...
        bucket = self.client.bucket(testrun_sibs_bucket)
        obj = bucket.get(self.key_name)
        bucket.allow_mult = True

        # Even if it previously existed, let's store a base resolved version
        # from which we can diverge by sending a stale vclock.
        obj.data = 'start'
        obj.content_type = 'text/plain'
        obj.store()

        vals = set(self.generate_siblings(obj, count=5))

        # Make sure the object has five siblings...
        obj = bucket.get(self.key_name)
        obj.reload()
        self.assertEqual(len(obj.siblings), 5)

        # When the object is in conflict, using the shortcut methods
        # should raise the ConflictError
        with self.assertRaises(ConflictError):
            obj.data

        # Get each of the values - make sure they match what was
        # assigned
        vals2 = set([sibling.data for sibling in obj.siblings])
        self.assertEqual(vals, vals2)

        # Resolve the conflict, and then do a get...
        resolved_sibling = obj.siblings[3]
        obj.siblings = [resolved_sibling]
        obj.store()

        obj.reload()
        self.assertEqual(len(obj.siblings), 1)
        self.assertEqual(obj.data, resolved_sibling.data)

    @unittest.skipUnless(RUN_RESOLVE, "RUN_RESOLVE is 0")
    def test_resolution(self):
        bucket = self.client.bucket(testrun_sibs_bucket)
        obj = bucket.get(self.key_name)
        bucket.allow_mult = True

        # Even if it previously existed, let's store a base resolved version
        # from which we can diverge by sending a stale vclock.
        obj.data = 'start'
        obj.content_type = 'text/plain'
        obj.store()

        vals = self.generate_siblings(obj, count=5, delay=1.01)

        # Make sure the object has five siblings when using the
        # default resolver
        obj = bucket.get(self.key_name)
        obj.reload()
        self.assertEqual(len(obj.siblings), 5)

        # Setting the resolver on the client object to use the
        # "last-write-wins" behavior
        self.client.resolver = last_written_resolver
        obj.reload()
        self.assertEqual(obj.resolver, last_written_resolver)
        self.assertEqual(1, len(obj.siblings))
        self.assertEqual(obj.data, vals[-1])

        # Set the resolver on the bucket to the default resolver,
        # overriding the resolver on the client
        bucket.resolver = default_resolver
        obj.reload()
        self.assertEqual(obj.resolver, default_resolver)
        self.assertEqual(len(obj.siblings), 5)

        # Define our own custom resolver on the object that returns
        # the maximum value, overriding the bucket and client resolvers
        def max_value_resolver(obj):
            obj.siblings = [max(obj.siblings, key=lambda s: s.data), ]

        obj.resolver = max_value_resolver
        obj.reload()
        self.assertEqual(obj.resolver, max_value_resolver)
        self.assertEqual(obj.data, max(vals))

        # Setting the resolver to None on all levels reverts to the
        # default resolver.
        obj.resolver = None
        self.assertEqual(obj.resolver, default_resolver)  # set by bucket
        bucket.resolver = None
        self.assertEqual(obj.resolver, last_written_resolver)  # set by client
        self.client.resolver = None
        self.assertEqual(obj.resolver, default_resolver)  # reset
        self.assertEqual(bucket.resolver, default_resolver)  # reset
        self.assertEqual(self.client.resolver, default_resolver)  # reset

    @unittest.skipUnless(RUN_RESOLVE, "RUN_RESOLVE is 0")
    def test_resolution_default(self):
        # If no resolver is setup, be sure to resolve to default_resolver
        bucket = self.client.bucket(testrun_sibs_bucket)
        self.assertEqual(self.client.resolver, default_resolver)
        self.assertEqual(bucket.resolver, default_resolver)

    def test_tombstone_siblings(self):
        # Set up the bucket, clear any existing object...
        bucket = self.client.bucket(testrun_sibs_bucket)
        obj = bucket.get(self.key_name)
        bucket.allow_mult = True

        obj.data = 'start'
        obj.content_type = 'text/plain'
        obj.store(return_body=True)

        obj.delete()

        vals = set(self.generate_siblings(obj, count=4))

        obj = bucket.get(self.key_name)

        # TODO this used to be 5, only
        siblen = len(obj.siblings)
        self.assertTrue(siblen == 4 or siblen == 5)

        non_tombstones = 0
        for sib in obj.siblings:
            if sib.exists:
                non_tombstones += 1
            self.assertTrue(not sib.exists or sib.data in vals)
        self.assertEqual(non_tombstones, 4)

    def test_store_of_missing_object(self):
        bucket = self.client.bucket(self.bucket_name)
        # for json objects
        o = bucket.get(self.key_name)
        self.assertEqual(o.exists, False)
        o.data = {"foo": "bar"}
        o.content_type = 'application/json'

        o = o.store()
        self.assertEqual(o.data, {"foo": "bar"})
        self.assertEqual(o.content_type, "application/json")
        o.delete()
        # for binary objects
        o = bucket.get(self.randname())
        self.assertEqual(o.exists, False)
        if PY2:
            o.encoded_data = "1234567890"
        else:
            o.encoded_data = "1234567890".encode()
        o.content_type = 'application/octet-stream'

        o = o.store()
        if PY2:
            self.assertEqual(o.encoded_data, "1234567890")
        else:
            self.assertEqual(o.encoded_data, "1234567890".encode())
        self.assertEqual(o.content_type, "application/octet-stream")
        o.delete()

    def test_store_metadata(self):
        bucket = self.client.bucket(self.bucket_name)
        rand = self.randint()
        obj = bucket.new(self.key_name, rand)
        obj.usermeta = {'custom': 'some metadata'}
        obj.store()
        obj = bucket.get(self.key_name)
        self.assertEqual('some metadata', obj.usermeta['custom'])

    def test_list_buckets(self):
        bucket = self.client.bucket(self.bucket_name)
        bucket.new("one", {"foo": "one", "bar": "red"}).store()
        buckets = self.client.get_buckets()
        self.assertTrue(self.bucket_name in [x.name for x in buckets])

    def test_stream_buckets(self):
        bucket = self.client.bucket(self.bucket_name)
        bucket.new(self.key_name, data={"foo": "one",
                                        "bar": "baz"}).store()
        buckets = []
        for bucket_list in self.client.stream_buckets():
            buckets.extend(bucket_list)

        self.assertTrue(self.bucket_name in [x.name for x in buckets])

    def test_stream_buckets_abort(self):
        bucket = self.client.bucket(self.bucket_name)
        bucket.new(self.key_name, data={"foo": "one",
                                        "bar": "baz"}).store()
        try:
            for bucket_list in self.client.stream_buckets():
                raise RuntimeError("abort")
        except RuntimeError:
            pass

        robj = bucket.get(self.key_name)
        self.assertTrue(robj.exists)
        self.assertEqual(len(robj.siblings), 1)

    def test_get_params(self):
        bucket = self.client.bucket(self.bucket_name)
        bucket.new(self.key_name, data={"foo": "one",
                                        "bar": "baz"}).store()

        bucket.get(self.key_name, basic_quorum=False)
        bucket.get(self.key_name, basic_quorum=True)
        bucket.get(self.key_name, notfound_ok=True)
        bucket.get(self.key_name, notfound_ok=False)

        missing = bucket.get('missing-key', notfound_ok=True,
                             basic_quorum=True)
        self.assertFalse(missing.exists)

    def test_preflist(self):
        bucket = self.client.bucket(self.bucket_name)
        bucket.new(self.key_name, data={"foo": "one",
                                        "bar": "baz"}).store()
        preflist = bucket.get_preflist(self.key_name)
        preflist2 = self.client.get_preflist(bucket, self.key_name)
        nodes = ['riak@127.0.0.1', 'dev1@127.0.0.1']
        for pref in (preflist, preflist2):
            self.assertEqual(len(pref), 3)
            self.assertIn(pref[0]['node'], nodes)
            [self.assertTrue(node['primary']) for node in pref]

    def generate_siblings(self, original, count=5, delay=None):
        vals = []
        for _ in range(count):
            while True:
                randval = str(self.randint())
                if randval not in vals:
                    break

            other_obj = original.bucket.new(key=original.key,
                                            data=randval,
                                            content_type='text/plain')
            other_obj.vclock = original.vclock
            other_obj.store()
            vals.append(randval)
            if delay:
                sleep(delay)
        return vals


class BucketPropsTest(IntegrationTestBase, unittest.TestCase):
    def test_rw_settings(self):
        bucket = self.client.bucket(testrun_props_bucket)
        self.assertEqual(bucket.r, "quorum")
        self.assertEqual(bucket.w, "quorum")
        self.assertEqual(bucket.dw, "quorum")
        self.assertEqual(bucket.rw, "quorum")

        bucket.w = 1
        self.assertEqual(bucket.w, 1)

        bucket.r = "quorum"
        self.assertEqual(bucket.r, "quorum")

        bucket.dw = "all"
        self.assertEqual(bucket.dw, "all")

        bucket.rw = "one"
        self.assertEqual(bucket.rw, "one")

        bucket.set_properties({'w': 'quorum',
                               'r': 'quorum',
                               'dw': 'quorum',
                               'rw': 'quorum'})
        bucket.clear_properties()

    def test_primary_quora(self):
        bucket = self.client.bucket(testrun_props_bucket)
        self.assertEqual(bucket.pr, 0)
        self.assertEqual(bucket.pw, 0)

        bucket.pr = 1
        self.assertEqual(bucket.pr, 1)

        bucket.pw = "quorum"
        self.assertEqual(bucket.pw, "quorum")

        bucket.set_properties({'pr': 0, 'pw': 0})
        bucket.clear_properties()

    def test_clear_bucket_properties(self):
        bucket = self.client.bucket(testrun_props_bucket)
        bucket.allow_mult = True
        self.assertTrue(bucket.allow_mult)
        bucket.n_val = 1
        self.assertEqual(bucket.n_val, 1)
        # Test setting clearing properties...

        self.assertTrue(bucket.clear_properties())
        self.assertFalse(bucket.allow_mult)
        self.assertEqual(bucket.n_val, 3)


class KVFileTests(IntegrationTestBase, unittest.TestCase):
    def test_store_binary_object_from_file(self):
        bucket = self.client.bucket(self.bucket_name)
        obj = bucket.new_from_file(self.key_name, __file__)
        obj.store()
        obj = bucket.get(self.key_name)
        self.assertNotEqual(obj.encoded_data, None)
        self.assertTrue(obj.content_type == 'text/x-python' or
                        obj.content_type == 'application/x-python-code')

    def test_store_binary_object_from_file_should_use_default_mimetype(self):
        bucket = self.client.bucket(self.bucket_name)
        filepath = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                os.pardir, os.pardir, 'README.md')
        obj = bucket.new_from_file(self.key_name, filepath)
        obj.store()
        obj = bucket.get(self.key_name)
        self.assertEqual(obj.content_type, 'application/octet-stream')

    def test_store_binary_object_from_file_should_fail_if_file_not_found(self):
        bucket = self.client.bucket(self.bucket_name)
        with self.assertRaises(IOError):
            bucket.new_from_file(self.key_name, 'FILE_NOT_FOUND')
        obj = bucket.get(self.key_name)
        # self.assertEqual(obj.encoded_data, None)
        self.assertFalse(obj.exists)


class CounterTests(IntegrationTestBase, unittest.TestCase):
    def test_counter_requires_allow_mult(self):
        bucket = self.client.bucket(self.bucket_name)
        if bucket.allow_mult:
            bucket.allow_mult = False
        self.assertFalse(bucket.allow_mult)

        with self.assertRaises(Exception):
            bucket.update_counter(self.key_name, 10)

    def test_counter_ops(self):
        bucket = self.client.bucket(testrun_sibs_bucket)
        self.assertTrue(bucket.allow_mult)

        # Non-existent counter has no value
        self.assertEqual(None, bucket.get_counter(self.key_name))

        # Update the counter
        bucket.update_counter(self.key_name, 10)
        self.assertEqual(10, bucket.get_counter(self.key_name))

        # Update with returning the value
        self.assertEqual(15, bucket.update_counter(self.key_name, 5,
                                                   returnvalue=True))

        # Now try decrementing
        self.assertEqual(10, bucket.update_counter(self.key_name, -5,
                                                   returnvalue=True))



"""
Copyright 2015 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import socket
from riak.security import SecurityError, USE_STDLIB_SSL
if USE_STDLIB_SSL:
    import ssl
else:
    import OpenSSL.SSL
    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO


def verify_cb(conn, cert, errnum, depth, ok):
    """
    The default OpenSSL certificate verification callback.
    """
    if not ok:
        raise SecurityError("Could not verify CA certificate {0}"
                            .format(cert.get_subject()))
    return ok


if USE_STDLIB_SSL:
    def configure_ssl_context(credentials):
        """
        Set various options on the SSL context for Python >= 2.7.9 and 3.x.

        N.B. versions earlier than 3.4 may not support all security
        measures, e.g., hostname check.

        :param credentials: Riak Security Credentials
        :type credentials: :class:`~riak.security.SecurityCreds`
        :rtype :class:`~ssl.SSLContext`
        """

        ssl_ctx = ssl.SSLContext(credentials.ssl_version)
        ssl_ctx.verify_mode = ssl.CERT_REQUIRED
        if hasattr(ssl_ctx, 'check_hostname'):
            ssl_ctx.check_hostname = True
        if credentials.cacert_file is None:
            raise SecurityError("cacert_file is required in SecurityCreds")
        if credentials.ciphers is not None:
            ssl_ctx.set_ciphers(credentials.ciphers)

        ssl_ctx.load_verify_locations(credentials.cacert_file)
        if credentials.ciphers is not None:
            ssl_ctx.set_ciphers(credentials.ciphers)

        pkeyfile = credentials.pkey_file
        certfile = credentials.cert_file
        if pkeyfile and not certfile:
            raise SecurityError("cert_file must be specified with pkey_file")
        if certfile and not pkeyfile:
            pkeyfile = certfile
        if certfile:
            ssl_ctx.load_cert_chain(certfile, pkeyfile)
        # TODO https://bugs.python.org/issue8813
        if credentials.crl_file is not None:
            ssl_ctx.load_verify_locations(credentials.crl_file)
            ssl_ctx.verify_flags = ssl.VERIFY_CRL_CHECK_LEAF

        # SSLv2 considered harmful.
        ssl_ctx.options |= ssl.OP_NO_SSLv2

        # SSLv3 has problematic security and is only required for really old
        # clients such as IE6 on Windows XP
        ssl_ctx.options |= ssl.OP_NO_SSLv3

        # disable compression to prevent CRIME attacks (OpenSSL 1.0+)
        ssl_ctx.options |= ssl.OP_NO_COMPRESSION

        return ssl_ctx

else:
    def configure_pyopenssl_context(credentials):
        """
        Set various options on the SSL context for Python <= 2.7.8.

        :param credentials: Riak Security Credentials
        :type credentials: :class:`~riak.security.SecurityCreds`
        :rtype ssl_ctx: :class:`~OpenSSL.SSL.Context`
        """

        ssl_ctx = OpenSSL.SSL.Context(credentials.ssl_version)
        if credentials._has_credential('pkey'):
            ssl_ctx.use_privatekey(credentials.pkey)
        if credentials._has_credential('cert'):
            ssl_ctx.use_certificate(credentials.cert)
        if credentials._has_credential('cacert'):
            store = ssl_ctx.get_cert_store()
            cacerts = credentials.cacert
            if not isinstance(cacerts, list):
                cacerts = [cacerts]
            for cacert in cacerts:
                store.add_cert(cacert)
        else:
            raise SecurityError("cacert_file is required in SecurityCreds")
        ciphers = credentials.ciphers
        if ciphers is not None:
            ssl_ctx.set_cipher_list(ciphers)
        # Demand a certificate
        ssl_ctx.set_verify(OpenSSL.SSL.VERIFY_PEER |
                           OpenSSL.SSL.VERIFY_FAIL_IF_NO_PEER_CERT,
                           verify_cb)
        return ssl_ctx

    # Inspired by
    # https://github.com/shazow/urllib3/blob/master/urllib3/contrib/pyopenssl.py
    class RiakWrappedSocket(socket.socket):
        def __init__(self, connection, socket):
            """
            API-compatibility wrapper for Python OpenSSL's Connection-class.

            :param connection: OpenSSL connection
            :type connection: OpenSSL.SSL.Connection
            :param socket: Underlying already connected socket
            :type socket: socket
            """
            self.connection = connection
            self.socket = socket

        def fileno(self):
            return self.socket.fileno()

        def makefile(self, mode, bufsize=-1):
            return fileobject(self.connection, mode, bufsize)

        def settimeout(self, timeout):
            return self.socket.settimeout(timeout)

        def sendall(self, data):
            # SSL seems to need bytes, so force the data to byte encoding
            return self.connection.sendall(bytes(data))

        def close(self):
            try:
                return self.connection.shutdown()
            except OpenSSL.SSL.Error as err:
                if err.args == ([],):
                    return False
                else:
                    raise err

    # Blatantly Stolen from
    # https://github.com/shazow/urllib3/blob/master/urllib3/contrib/pyopenssl.py
    # which is basically a port of the `socket._fileobject` class
    class fileobject(socket._fileobject):
        """
        Extension of the socket module's fileobject to use PyOpenSSL.
        """

        def read(self, size=-1):
            # Use max, disallow tiny reads in a loop as they are very
            # inefficient.  We never leave read() with any leftover data from
            # a new recv() call in our internal buffer.
            rbufsize = max(self._rbufsize, self.default_bufsize)
            # Our use of StringIO rather than lists of string objects returned
            # by recv() minimizes memory usage and fragmentation that occurs
            # when rbufsize is large compared to the typical return value of
            # recv().
            buf = self._rbuf
            buf.seek(0, 2)  # seek end
            if size < 0:
                # Read until EOF
                self._rbuf = StringIO()  # reset _rbuf.  we consume it via buf.
                while True:
                    try:
                        data = self._sock.recv(rbufsize)
                    except OpenSSL.SSL.WantReadError:
                        continue
                    if not data:
                        break
                    buf.write(data)
                return buf.getvalue()
            else:
                # Read until size bytes or EOF seen, whichever comes first
                buf_len = buf.tell()
                if buf_len >= size:
                    # Already have size bytes in our buffer?  Extract and
                    # return.
                    buf.seek(0)
                    rv = buf.read(size)
                    self._rbuf = StringIO()
                    self._rbuf.write(buf.read())
                    return rv

                self._rbuf = StringIO()  # reset _rbuf.  we consume it via buf.
                while True:
                    left = size - buf_len
                    # recv() will malloc the amount of memory given as its
                    # parameter even though it often returns much less data
                    # than that.  The returned data string is short lived
                    # as we copy it into a StringIO and free it.  This avoids
                    # fragmentation issues on many platforms.
                    try:
                        data = self._sock.recv(left)
                    except OpenSSL.SSL.WantReadError:
                        continue
                    if not data:
                        break
                    n = len(data)
                    if n == size and not buf_len:
                        # Shortcut.  Avoid buffer data copies when:
                        # - We have no data in our buffer.
                        # AND
                        # - Our call to recv returned exactly the
                        #   number of bytes we were asked to read.
                        return data
                    if n == left:
                        buf.write(data)
                        # del data  # explicit free
                        break
                    assert n <= left, "recv(%d) returned %d bytes" % (left, n)
                    buf.write(data)
                    buf_len += n
                    # del data  # explicit free
                    # assert buf_len == buf.tell()
                # Moved del outside of loop to keep pyflakes happy
                if data:
                    del data
                return buf.getvalue()

        def readline(self, size=-1):
            data = None
            buf = self._rbuf
            buf.seek(0, 2)  # seek end
            if buf.tell() > 0:
                # check if we already have it in our buffer
                buf.seek(0)
                bline = buf.readline(size)
                if bline.endswith('\n') or len(bline) == size:
                    self._rbuf = StringIO()
                    self._rbuf.write(buf.read())
                    return bline
                del bline
            if size < 0:
                # Read until \n or EOF, whichever comes first
                if self._rbufsize <= 1:
                    # Speed up unbuffered case
                    buf.seek(0)
                    buffers = [buf.read()]
                    # reset _rbuf.  we consume it via buf.
                    self._rbuf = StringIO()
                    data = None
                    recv = self._sock.recv
                    while True:
                        try:
                            while data != "\n":
                                data = recv(1)
                                if not data:
                                    break
                                buffers.append(data)
                        except OpenSSL.SSL.WantReadError:
                            continue
                        break
                    return "".join(buffers)

                buf.seek(0, 2)  # seek end
                self._rbuf = StringIO()  # reset _rbuf.  we consume it via buf.
                while True:
                    try:
                        data = self._sock.recv(self._rbufsize)
                    except OpenSSL.SSL.WantReadError:
                        continue
                    if not data:
                        break
                    nl = data.find('\n')
                    if nl >= 0:
                        nl += 1
                        buf.write(data[:nl])
                        self._rbuf.write(data[nl:])
                        # del data
                        break
                    buf.write(data)
                # Moved del outside of loop to keep pyflakes happy
                if data:
                    del data
                return buf.getvalue()
            else:
                # Read until size bytes or \n or EOF seen, whichever comes 1st
                buf.seek(0, 2)  # seek end
                buf_len = buf.tell()
                if buf_len >= size:
                    buf.seek(0)
                    rv = buf.read(size)
                    self._rbuf = StringIO()
                    self._rbuf.write(buf.read())
                    return rv
                self._rbuf = StringIO()  # reset _rbuf.  we consume it via buf.
                while True:
                    try:
                        data = self._sock.recv(self._rbufsize)
                    except OpenSSL.SSL.WantReadError:
                            continue
                    if not data:
                        break
                    left = size - buf_len
                    # did we just receive a newline?
                    nl = data.find('\n', 0, left)
                    if nl >= 0:
                        nl += 1
                        # save the excess data to _rbuf
                        self._rbuf.write(data[nl:])
                        if buf_len:
                            buf.write(data[:nl])
                            break
                        else:
                            # Shortcut.  Avoid data copy through buf when
                            # returning a substring of our first recv().
                            return data[:nl]
                    n = len(data)
                    if n == size and not buf_len:
                        # Shortcut.  Avoid data copy through buf when
                        # returning exactly all of our first recv().
                        return data
                    if n >= left:
                        buf.write(data[:left])
                        self._rbuf.write(data[left:])
                        break
                    buf.write(data)
                    buf_len += n
                    # assert buf_len == buf.tell()
                return buf.getvalue()



# -*- coding: utf-8 -*-
import platform
from riak.tests import RUN_YZ
from riak.tests.base import IntegrationTestBase
from riak.tests.comparison import Comparison
from riak.tests.yz_setup import yzSetUp, yzTearDown

if platform.python_version() < '2.7':
    unittest = __import__('unittest2')
else:
    import unittest


def wait_for_yz_index(bucket, key, index=None):
    """
    Wait until Solr index has been updated and a value returns from a query.

    :param bucket: Bucket to which indexed value is written
    :type bucket: RiakBucket
    :param key: Key to which value was written
    :type key: str
    """
    while len(bucket.search('_yz_rk:' + key, index=index)['docs']) == 0:
        pass

# YZ index on bucket of the same name
testrun_yz = {'btype': None, 'bucket': 'yzbucket', 'index': 'yzbucket'}
# YZ index on bucket of a different name
testrun_yz_index = {'btype': None,
                    'bucket': 'yzindexbucket',
                    'index': 'yzindex'}


def setUpModule():
    yzSetUp(testrun_yz, testrun_yz_index)


def tearDownModule():
    yzTearDown(testrun_yz, testrun_yz_index)


@unittest.skipUnless(RUN_YZ, 'RUN_YZ is 0')
class YZSearchTests(IntegrationTestBase, unittest.TestCase, Comparison):
    def test_yz_search_from_bucket(self):
        bucket = self.client.bucket(testrun_yz['bucket'])
        bucket.new("user", {"user_s": "Z"}).store()
        wait_for_yz_index(bucket, "user")
        results = bucket.search("user_s:Z")
        self.assertEqual(1, len(results['docs']))
        # TODO: check that docs return useful info
        result = results['docs'][0]
        self.assertIn('_yz_rk', result)
        self.assertEqual(u'user', result['_yz_rk'])
        self.assertIn('_yz_rb', result)
        self.assertEqual(testrun_yz['bucket'], result['_yz_rb'])
        self.assertIn('score', result)
        self.assertIn('user_s', result)
        self.assertEqual(u'Z', result['user_s'])

    def test_yz_search_index_using_bucket(self):
        bucket = self.client.bucket(testrun_yz_index['bucket'])
        bucket.new("feliz",
                   {"name_s": "Felix", "species_s": "Felis catus"}).store()
        wait_for_yz_index(bucket, "feliz", index=testrun_yz_index['index'])
        results = bucket.search('name_s:Felix',
                                index=testrun_yz_index['index'])
        self.assertEqual(1, len(results['docs']))

    def test_yz_search_index_using_wrong_bucket(self):
        bucket = self.client.bucket(testrun_yz_index['bucket'])
        bucket.new("feliz",
                   {"name_s": "Felix", "species_s": "Felis catus"}).store()
        wait_for_yz_index(bucket, "feliz", index=testrun_yz_index['index'])
        with self.assertRaises(Exception):
            bucket.search('name_s:Felix')

    def test_yz_get_search_index(self):
        index = self.client.get_search_index(testrun_yz['bucket'])
        self.assertEqual(testrun_yz['bucket'], index['name'])
        self.assertEqual('_yz_default', index['schema'])
        self.assertEqual(3, index['n_val'])
        with self.assertRaises(Exception):
            self.client.get_search_index('NOT' + testrun_yz['bucket'])

    def test_yz_delete_search_index(self):
        # expected to fail, since there's an attached bucket
        with self.assertRaises(Exception):
            self.client.delete_search_index(testrun_yz['bucket'])
        # detatch bucket from index then delete
        b = self.client.bucket(testrun_yz['bucket'])
        b.set_property('search_index', '_dont_index_')
        self.assertTrue(self.client.delete_search_index(testrun_yz['bucket']))
        # create it again
        self.client.create_search_index(testrun_yz['bucket'], '_yz_default', 3)
        b = self.client.bucket(testrun_yz['bucket'])
        b.set_property('search_index', testrun_yz['bucket'])
        # Wait for index to apply
        indexes = []
        while testrun_yz['bucket'] not in indexes:
            indexes = [i['name'] for i in self.client.list_search_indexes()]

    def test_yz_list_search_indexes(self):
        indexes = self.client.list_search_indexes()
        self.assertIn(testrun_yz['bucket'], [item['name'] for item in indexes])
        self.assertLessEqual(1, len(indexes))

    def test_yz_create_schema(self):
        content = """<?xml version="1.0" encoding="UTF-8" ?>
        <schema name="test" version="1.5">
        <fields>
           <field name="_yz_id" type="_yz_str" indexed="true" stored="true"
            multiValued="false" required="true" />
           <field name="_yz_ed" type="_yz_str" indexed="true" stored="true"
            multiValued="false" />
           <field name="_yz_pn" type="_yz_str" indexed="true" stored="true"
            multiValued="false" />
           <field name="_yz_fpn" type="_yz_str" indexed="true" stored="true"
            multiValued="false" />
           <field name="_yz_vtag" type="_yz_str" indexed="true" stored="true"
            multiValued="false" />
           <field name="_yz_rk" type="_yz_str" indexed="true" stored="true"
            multiValued="false" />
           <field name="_yz_rb" type="_yz_str" indexed="true" stored="true"
            multiValued="false" />
           <field name="_yz_rt" type="_yz_str" indexed="true" stored="true"
            multiValued="false" />
           <field name="_yz_err" type="_yz_str" indexed="true"
            multiValued="false" />
        </fields>
        <uniqueKey>_yz_id</uniqueKey>
        <types>
            <fieldType name="_yz_str" class="solr.StrField"
             sortMissingLast="true" />
        </types>
        </schema>"""
        schema_name = self.randname()
        self.assertTrue(self.client.create_search_schema(schema_name, content))
        schema = self.client.get_search_schema(schema_name)
        self.assertEqual(schema_name, schema['name'])
        self.assertEqual(content, schema['content'])

    def test_yz_create_bad_schema(self):
        bad_content = """
        <derp nope nope, how do i computer?
        """
        with self.assertRaises(Exception):
            self.client.create_search_schema(self.randname(), bad_content)

    def test_yz_search_queries(self):
        bucket = self.client.bucket(testrun_yz['bucket'])
        bucket.new("Z", {"username_s": "Z", "name_s": "ryan",
                         "age_i": 30}).store()
        bucket.new("R", {"username_s": "R", "name_s": "eric",
                         "age_i": 34}).store()
        bucket.new("F", {"username_s": "F", "name_s": "bryan fink",
                         "age_i": 32}).store()
        bucket.new("H", {"username_s": "H", "name_s": "brett",
                         "age_i": 14}).store()
        wait_for_yz_index(bucket, "H")
        # multiterm
        results = bucket.search("username_s:(F OR H)")
        self.assertEqual(2, len(results['docs']))
        # boolean
        results = bucket.search("username_s:Z AND name_s:ryan")
        self.assertEqual(1, len(results['docs']))
        # range
        results = bucket.search("age_i:[30 TO 33]")
        self.assertEqual(2, len(results['docs']))
        # phrase
        results = bucket.search('name_s:"bryan fink"')
        self.assertEqual(1, len(results['docs']))
        # wildcard
        results = bucket.search('name_s:*ryan*')
        self.assertEqual(2, len(results['docs']))
        # regexp
        results = bucket.search('name_s:/br.*/')
        self.assertEqual(2, len(results['docs']))
        # Parameters:
        # limit
        results = bucket.search('username_s:*', rows=2)
        self.assertEqual(2, len(results['docs']))
        # sort
        results = bucket.search('username_s:*', sort="age_i asc")
        self.assertEqual(14, int(results['docs'][0]['age_i']))

    def test_yz_search_utf8(self):
        bucket = self.client.bucket(testrun_yz['bucket'])
        body = {"text_ja": u"私はハイビスカスを食べるのが 大好き"}
        bucket.new(self.key_name, body).store()
        wait_for_yz_index(bucket, self.key_name)
        results = bucket.search(u"text_ja:大好き AND  _yz_rk:{0}".
                                format(self.key_name))
        self.assertEqual(1, len(results['docs']))

    def test_yz_multivalued_fields(self):
        bucket = self.client.bucket(testrun_yz['bucket'])
        body = {"groups_ss": ['a', 'b', 'c']}
        bucket.new(self.key_name, body).store()
        wait_for_yz_index(bucket, self.key_name)
        results = bucket.search('groups_ss:* AND _yz_rk:{0}'.
                                format(self.key_name))
        self.assertEqual(1, len(results['docs']))
        doc = results['docs'][0]
        self.assertIn('groups_ss', doc)
        field = doc['groups_ss']
        self.assertIsInstance(field, list)
        self.assertItemsEqual(['a', 'b', 'c'], field)



from six import string_types, PY2


class Table(object):
    """
    The ``Table`` object allows you to access properties on a Riak
    timeseries table and query timeseries data.
    """
    def __init__(self, client, name):
        """
        Returns a new ``Table`` instance.

        :param client: A :class:`RiakClient <riak.client.RiakClient>`
               instance
        :type client: :class:`RiakClient <riak.client.RiakClient>`
        :param name: The table's name
        :type name: string
        """
        if not isinstance(name, string_types):
            raise TypeError('Table name must be a string')

        if PY2:
            try:
                name = name.encode('ascii')
            except UnicodeError:
                raise TypeError('Unicode table names are not supported.')

        self._client = client
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def new(self, rows, columns=None):
        """
        A shortcut for manually instantiating a new
        :class:`~riak.ts_object.TsObject`

        :param rows: An list of lists with timeseries data
        :type rows: list
        :param columns: An list of Column names and types. Optional.
        :type columns: list
        :rtype: :class:`~riak.ts_object.TsObject`
        """
        from riak.ts_object import TsObject

        return TsObject(self._client, self, rows, columns)

    def describe(self):
        """
        Retrieves a timeseries table's description.

        :rtype: :class:`TsObject <riak.ts_object.TsObject>`
        """
        return self._client.ts_describe(self)

    def get(self, key):
        """
        Gets a value from a timeseries table.

        :param key: The timeseries value's key.
        :type key: list
        :rtype: :class:`TsObject <riak.ts_object.TsObject>`
        """
        return self._client.ts_get(self, key)

    def delete(self, key):
        """
        Deletes a value from a timeseries table.

        :param key: The timeseries value's key.
        :type key: list or dict
        :rtype: boolean
        """
        return self._client.ts_delete(self, key)

    def query(self, query, interpolations=None):
        """
        Queries a timeseries table.

        :param query: The timeseries query.
        :type query: string
        :rtype: :class:`TsObject <riak.ts_object.TsObject>`
        """
        return self._client.ts_query(self, query, interpolations)

    def stream_keys(self, timeout=None):
        """
        Streams keys from a timeseries table.

        :rtype: list
        """
        return self._client.ts_stream_keys(self, timeout)



class XMLSearchResult(object):
    # Match tags that are document fields
    fieldtags = ['str', 'int', 'date']

    def __init__(self):
        # Results
        self.num_found = 0
        self.max_score = 0.0
        self.docs = []

        # Parser state
        self.currdoc = None
        self.currfield = None
        self.currvalue = None

    def start(self, tag, attrib):
        if tag == 'result':
            self.num_found = int(attrib['numFound'])
            self.max_score = float(attrib['maxScore'])
        elif tag == 'doc':
            self.currdoc = {}
        elif tag in self.fieldtags and self.currdoc is not None:
            self.currfield = attrib['name']

    def end(self, tag):
        if tag == 'doc' and self.currdoc is not None:
            self.docs.append(self.currdoc)
            self.currdoc = None
        elif tag in self.fieldtags and self.currdoc is not None:
            if tag == 'int':
                self.currvalue = int(self.currvalue)
            self.currdoc[self.currfield] = self.currvalue
            self.currfield = None
            self.currvalue = None

    def data(self, data):
        if self.currfield:
            # riak_solr_output adds NL + 6 spaces
            data = data.rstrip()
            if self.currvalue:
                self.currvalue += data
            else:
                self.currvalue = data

    def close(self):
        return {'num_found': self.num_found,
                'max_score': self.max_score,
                'docs': self.docs}



# -*- coding: utf-8 -*-
from six import PY2, PY3
import collections
import warnings


class Comparison(object):
    '''
    Provide a cross-version object comparison operator
    since its name changed between Python 2.x and Python 3.x
    '''

    if PY3:
        # Stolen from Python 2.7.8's unittest
        _Mismatch = collections.namedtuple('Mismatch', 'actual expected value')

        def _count_diff_all_purpose(self, actual, expected):
            '''
            Returns list of (cnt_act, cnt_exp, elem)
            triples where the counts differ
            '''
            # elements need not be hashable
            s, t = list(actual), list(expected)
            m, n = len(s), len(t)
            NULL = object()
            result = []
            for i, elem in enumerate(s):
                if elem is NULL:
                    continue
                cnt_s = cnt_t = 0
                for j in range(i, m):
                    if s[j] == elem:
                        cnt_s += 1
                        s[j] = NULL
                for j, other_elem in enumerate(t):
                    if other_elem == elem:
                        cnt_t += 1
                        t[j] = NULL
                if cnt_s != cnt_t:
                    diff = self._Mismatch(cnt_s, cnt_t, elem)
                    result.append(diff)

            for i, elem in enumerate(t):
                if elem is NULL:
                    continue
                cnt_t = 0
                for j in range(i, n):
                    if t[j] == elem:
                        cnt_t += 1
                        t[j] = NULL
                diff = self._Mismatch(0, cnt_t, elem)
                result.append(diff)
            return result

        def _count_diff_hashable(self, actual, expected):
            '''
            Returns list of (cnt_act, cnt_exp, elem) triples
            where the counts differ
            '''
            # elements must be hashable
            s, t = self._ordered_count(actual), self._ordered_count(expected)
            result = []
            for elem, cnt_s in s.items():
                cnt_t = t.get(elem, 0)
                if cnt_s != cnt_t:
                    diff = self._Mismatch(cnt_s, cnt_t, elem)
                    result.append(diff)
            for elem, cnt_t in t.items():
                if elem not in s:
                    diff = self._Mismatch(0, cnt_t, elem)
                    result.append(diff)
            return result

        def _ordered_count(self, iterable):
            'Return dict of element counts, in the order they were first seen'
            c = collections.OrderedDict()
            for elem in iterable:
                c[elem] = c.get(elem, 0) + 1
            return c

        def assertItemsEqual(self, expected_seq, actual_seq, msg=None):
            """An unordered sequence specific comparison. It asserts that
            actual_seq and expected_seq have the same element counts.
            Equivalent to::

                self.assertEqual(Counter(iter(actual_seq)),
                                 Counter(iter(expected_seq)))

            Asserts that each element has the same count in both sequences.
            Example:
                - [0, 1, 1] and [1, 0, 1] compare equal.
                - [0, 0, 1] and [0, 1] compare unequal.
            """
            first_seq, second_seq = list(expected_seq), list(actual_seq)
            with warnings.catch_warnings():
                try:
                    first = collections.Counter(first_seq)
                    second = collections.Counter(second_seq)
                except TypeError:
                    # Handle case with unhashable elements
                    differences = self._count_diff_all_purpose(first_seq,
                                                               second_seq)
                else:
                    if first == second:
                        return
                    differences = self._count_diff_hashable(first_seq,
                                                            second_seq)

            if differences:
                standardMsg = 'Element counts were not equal:\n'
                lines = ['First has %d, Second has %d:  %r' %
                         diff for diff in differences]
                diffMsg = '\n'.join(lines)
                standardMsg = self._truncateMessage(standardMsg, diffMsg)

    def assert_raises_regex(self, exception, regexp):
        if PY2:
            return self.assertRaisesRegexp(exception, regexp)
        else:
            return self.assertRaisesRegex(exception, regexp)



"""
Copyright 2013 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from __future__ import print_function
import os
import gc

__all__ = ['measure', 'measure_with_rehearsal']


def measure_with_rehearsal():
    """
    Runs a benchmark when used as an iterator, injecting a garbage
    collection between iterations. Example::

        for b in riak.benchmark.measure_with_rehearsal():
            with b.report("pow"):
                for _ in range(10000):
                    math.pow(2,10000)
            with b.report("factorial"):
                for i in range(100):
                    math.factorial(i)
    """
    return Benchmark(True)


def measure():
    """
    Runs a benchmark once when used as a context manager. Example::

        with riak.benchmark.measure() as b:
            with b.report("pow"):
                for _ in range(10000):
                    math.pow(2,10000)
            with b.report("factorial"):
                for i in range(100):
                    math.factorial(i)
    """
    return Benchmark()


class Benchmark(object):
    """
    A benchmarking run, which may consist of multiple steps. See
    measure_with_rehearsal() and measure() for examples.
    """
    def __init__(self, rehearse=False):
        """
        Creates a new benchmark reporter.

        :param rehearse: whether to run twice to take counter the effects
           of garbage collection
        :type rehearse: boolean
        """
        self.rehearse = rehearse
        if rehearse:
            self.count = 2
        else:
            self.count = 1
        self._report = None

    def __enter__(self):
        if self.rehearse:
            raise ValueError("measure_with_rehearsal() cannot be used in with "
                             "statements, use measure() or the for..in "
                             "statement")
        print_header()
        self._report = BenchmarkReport()
        self._report.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._report:
            return self._report.__exit__(exc_type, exc_val, exc_tb)
        else:
            print
            return True

    def __iter__(self):
        return self

    def next(self):
        """
        Runs the next iteration of the benchmark.
        """
        if self.count == 0:
            raise StopIteration
        elif self.count > 1:
            print_rehearsal_header()
        else:
            if self.rehearse:
                gc.collect()
                print("-" * 59)
                print()
            print_header()

        self.count -= 1
        return self

    def __next__(self):
        # Python 3.x Version
        return self.next()

    def report(self, name):
        """
        Returns a report for the current step of the benchmark.
        """
        self._report = None
        return BenchmarkReport(name)


def print_rehearsal_header():
    """
    Prints the header for the rehearsal phase of a benchmark.
    """
    print
    print("Rehearsal -------------------------------------------------")


def print_report(label, user, system, real):
    """
    Prints the report of one step of a benchmark.
    """
    print("{:<12s} {:12f} {:12f} ( {:12f} )".format(label,
                                                    user,
                                                    system,
                                                    real))


def print_header():
    """
    Prints the header for the normal phase of a benchmark.
    """
    print("{:<12s} {:<12s} {:<12s} ( {:<12s} )"
          .format('', 'user', 'system', 'real'))


class BenchmarkReport(object):
    """
    A labeled step in a benchmark. Acts as a context-manager, printing
    its timing results when the context exits.
    """
    def __init__(self, name='benchmark'):
        self.name = name
        self.start = None

    def __enter__(self):
        self.start = os.times()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type:
            user1, system1, _, _, real1 = self.start
            user2, system2, _, _, real2 = os.times()
            print_report(self.name, user2 - user1, system2 - system1,
                         real2 - real1)
        elif exc_type is KeyboardInterrupt:
            return False
        else:
            print("EXCEPTION! %r" % ((exc_type, exc_val, exc_tb),))
        return True



"""
Copyright 2014 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import ssl
import warnings
from riak import RiakError
from riak.util import str_to_long

if hasattr(ssl, 'SSLContext'):
    # For Python >= 2.7.9 and Python 3.x
    USE_STDLIB_SSL = True
else:
    # For Python 2.6 and <= 2.7.8
    USE_STDLIB_SSL = False

if not USE_STDLIB_SSL:
    import OpenSSL.SSL
    from OpenSSL import crypto

OPENSSL_VERSION_101G = 268439679
if hasattr(ssl, 'OPENSSL_VERSION_NUMBER'):
    # For Python 2.7 and Python 3.x
    sslver = ssl.OPENSSL_VERSION_NUMBER
    # Be sure to use at least OpenSSL 1.0.1g
    tls_12 = hasattr(ssl, 'PROTOCOL_TLSv1_2')
    if sslver < OPENSSL_VERSION_101G or not tls_12:
        verstring = ssl.OPENSSL_VERSION
        msg = "{0} (>= 1.0.1g required), TLS 1.2 support: {1}" \
            .format(verstring, tls_12)
        warnings.warn(msg, UserWarning)
    if hasattr(ssl, 'PROTOCOL_TLSv1_2'):
        DEFAULT_TLS_VERSION = ssl.PROTOCOL_TLSv1_2
    elif hasattr(ssl, 'PROTOCOL_TLSv1_1'):
        DEFAULT_TLS_VERSION = ssl.PROTOCOL_TLSv1_1
    elif hasattr(ssl, 'PROTOCOL_TLSv1'):
        DEFAULT_TLS_VERSION = ssl.PROTOCOL_TLSv1
    else:
        DEFAULT_TLS_VERSION = ssl.PROTOCOL_SSLv23

else:
    # For Python 2.6
    sslver = OpenSSL.SSL.OPENSSL_VERSION_NUMBER
    # Be sure to use at least OpenSSL 1.0.1g
    tls_12 = hasattr(OpenSSL.SSL, 'TLSv1_2_METHOD')
    if (sslver < OPENSSL_VERSION_101G) or tls_12:
        verstring = OpenSSL.SSL.SSLeay_version(OpenSSL.SSL.SSLEAY_VERSION)
        msg = "{0} (>= 1.0.1g required), TLS 1.2 support: {1}" \
            .format(verstring, tls_12)
        warnings.warn(msg, UserWarning)
    if hasattr(OpenSSL.SSL, 'TLSv1_2_METHOD'):
        DEFAULT_TLS_VERSION = OpenSSL.SSL.TLSv1_2_METHOD
    elif hasattr(OpenSSL.SSL, 'TLSv1_1_METHOD'):
        DEFAULT_TLS_VERSION = OpenSSL.SSL.TLSv1_1_METHOD
    elif hasattr(OpenSSL.SSL, 'TLSv1_METHOD'):
        DEFAULT_TLS_VERSION = OpenSSL.SSL.TLSv1_METHOD
    else:
        DEFAULT_TLS_VERSION = OpenSSL.SSL.SSLv23_METHOD


class SecurityError(RiakError):
    """
    Raised when there is an issue establishing security.
    """
    def __init__(self, message="Security error"):
        super(SecurityError, self).__init__(message)


class SecurityCreds:
    def __init__(self,
                 username=None,
                 password=None,
                 pkey_file=None,
                 pkey=None,
                 cert_file=None,
                 cert=None,
                 cacert_file=None,
                 cacert=None,
                 crl_file=None,
                 crl=None,
                 ciphers=None,
                 ssl_version=DEFAULT_TLS_VERSION):
        """
        Container class for security-related settings

        :param username: Riak Security username
        :type username: str
        :param password: Riak Security password
        :type password: str
        :param pkey_file: Full path to security key file
        :type pkey_file: str
        :param key: Loaded security key file
        :type key: :class:`OpenSSL.crypto.PKey`
        :param cert_file: Full path to certificate file
        :type cert_file: str
        :param cert: Loaded client certificate
        :type cert: :class:`OpenSSL.crypto.X509`
        :param cacert_file: Full path to CA certificate file
        :type cacert_file: str
        :param cacert: Loaded CA certificate
        :type cacert: :class:`OpenSSL.crypto.X509`
        :param crl_file: Full path to revoked certificates file
        :type crl_file: str
        :param crl: Loaded revoked certificates list
        :type crl: :class:`OpenSSL.crypto.CRL`
        :param ciphers: List of supported SSL ciphers
        :type ciphers: str
        :param ssl_version: OpenSSL security version
        :type ssl_version: int
        """
        self._username = username
        self._password = password
        self._pkey_file = pkey_file
        self._pkey = pkey
        self._cert_file = cert_file
        self._cert = cert
        self._cacert_file = cacert_file
        self._cacert = cacert
        self._crl_file = crl_file
        self._crl = crl
        self._ciphers = ciphers
        self._ssl_version = ssl_version

    @property
    def username(self):
        """
        Riak Username

        :rtype: str
        """
        return self._username

    @property
    def password(self):
        """
        Riak Password

        :rtype: str
        """
        return self._password

    @property
    def pkey_file(self):
        """
        Client Private Key file

        :rtype: str
        """
        return self._pkey_file

    @property
    def cert_file(self):
        """
        Client Certificate file

        :rtype: str
        """
        return self._cert_file

    @property
    def cacert_file(self):
        """
        Certifying Authority (CA) Certificate file

        :rtype: str
        """
        return self._cacert_file

    @property
    def crl_file(self):
        """
        Certificate Revocation List file

        :rtype: str
        """
        return self._crl_file

    @property
    def ciphers(self):
        """
        Colon-delimited list of supported ciphers

        :rtype: str
        """
        return self._ciphers

    @property
    def ssl_version(self):
        """SSL/TLS Protocol to use

        :rtype: an int constant from OpenSSL, like
            :data:`OpenSSL.SSL.TLSv1_2_METHOD`
        """
        return self._ssl_version

    if not USE_STDLIB_SSL:
        @property
        def pkey(self):
            """
            Client Private key

            :rtype: :class:`OpenSSL.crypto.PKey`
            """
            return self._cached_cert('_pkey', crypto.load_privatekey)

        @property
        def cert(self):
            """
            Client Certificate

            :rtype: :class:`OpenSSL.crypto.X509`
            """
            return self._cached_cert('_cert', crypto.load_certificate)

        @property
        def cacert(self):
            """
            Certifying Authority (CA) Certificate

            :rtype: :class:`OpenSSL.crypto.X509`
            """
            return self._cached_cert('_cacert', crypto.load_certificate)

        @property
        def crl(self):
            """
            Certificate Revocation List

            :rtype: :class:`OpenSSL.crypto.CRL`
            """
            return self._cached_cert('_crl', crypto.load_crl)

        def _cached_cert(self, key, loader):
            # If the key is associated with a file,
            # then lazily load and cache it
            key_file = getattr(self, key + "_file")
            if (getattr(self, key) is None) and (key_file is not None):
                cert_list = []
                # The _file may be a list of files
                if not isinstance(key_file, list):
                    key_file = [key_file]
                for filename in key_file:
                    with open(filename, 'r') as f:
                        cert_list.append(loader(OpenSSL.SSL.FILETYPE_PEM,
                                                f.read()))
                # If it is not a list, just store the first element
                if len(cert_list) == 1:
                    cert_list = cert_list[0]
                setattr(self, key, cert_list)
            return getattr(self, key)

        def _has_credential(self, key):
            """
            ``True`` if a credential or filename value has been supplied for
            the given property.

            :param key: which configuration property to check for
            :type key: str
            :rtype: bool
            """
            internal_key = "_" + key
            return (getattr(self, internal_key) is not None) or \
                (getattr(self, internal_key + "_file") is not None)

        def _check_revoked_cert(self, ssl_socket):
            """
            Checks whether the server certificate on the passed socket has been
            revoked by checking the CRL.

            :param ssl_socket: the SSL/TLS socket
            :rtype: bool
            :raises SecurityError: when the certificate has been revoked
            """
            if not self._has_credential('crl'):
                return True

            servcert = ssl_socket.get_peer_certificate()
            servserial = servcert.get_serial_number()
            for rev in self.crl.get_revoked():
                if servserial == str_to_long(rev.get_serial(), 16):
                    raise SecurityError("Server certificate has been revoked")



"""
Copyright 2012 Basho Technologies, Inc.

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
from contextlib import contextmanager
from riak.transports.pool import BadResource
from riak.transports.pbc import is_retryable as is_pbc_retryable
from riak.transports.http import is_retryable as is_http_retryable
import threading
from six import PY2
if PY2:
    from httplib import HTTPException
else:
    from http.client import HTTPException

#: The default (global) number of times to retry requests that are
#: retryable. This can be modified locally, per-thread, via the
#: :attr:`RiakClient.retries` property, or using the
#: :attr:`RiakClient.retry_count` method in a ``with`` statement.
DEFAULT_RETRY_COUNT = 3


class _client_locals(threading.local):
    """
    A thread-locals object used by the client.
    """
    def __init__(self):
        self.riak_retries_count = DEFAULT_RETRY_COUNT


class RiakClientTransport(object):
    """
    Methods for RiakClient related to transport selection and retries.
    """

    # These will be set or redefined by the RiakClient initializer
    protocol = 'pbc'
    _http_pool = None
    _pb_pool = None
    _locals = _client_locals()

    def _get_retry_count(self):
        return self._locals.riak_retries_count or DEFAULT_RETRY_COUNT

    def _set_retry_count(self, value):
        if not isinstance(value, int):
            raise TypeError("retries must be an integer")
        self._locals.riak_retries_count = value

    __retries_doc = """
          The number of times retryable operations will be attempted
          before raising an exception to the caller. Defaults to
          ``3``.

          :note: This is a thread-local for safety and
                 operation-specific modification. To change the
                 default globally, modify
                 :data:`riak.client.transport.DEFAULT_RETRY_COUNT`.
          """

    retries = property(_get_retry_count, _set_retry_count, doc=__retries_doc)

    @contextmanager
    def retry_count(self, retries):
        """
        retry_count(retries)

        Modifies the number of retries for the scope of the ``with``
        statement (in the current thread).

        Example::

            with client.retry_count(10):
                client.ping()
        """
        if not isinstance(retries, int):
            raise TypeError("retries must be an integer")

        old_retries, self.retries = self.retries, retries
        try:
            yield
        finally:
            self.retries = old_retries

    @contextmanager
    def _transport(self):
        """
        _transport()

        Yields a single transport to the caller from the default pool,
        without retries.
        """
        pool = self._choose_pool()
        with pool.transaction() as transport:
            yield transport

    def _acquire(self):
        """
        _acquire()

        Acquires a connection from the default pool.
        """
        return self._choose_pool().acquire()

    def _with_retries(self, pool, fn):
        """
        Performs the passed function with retries against the given pool.

        :param pool: the connection pool to use
        :type pool: Pool
        :param fn: the function to pass a transport
        :type fn: function
        """
        skip_nodes = []

        def _skip_bad_nodes(transport):
            return transport._node not in skip_nodes

        retry_count = self.retries

        for retry in range(retry_count):
            try:
                with pool.transaction(_filter=_skip_bad_nodes) as transport:
                    try:
                        return fn(transport)
                    except (IOError, HTTPException) as e:
                        if _is_retryable(e):
                            transport._node.error_rate.incr(1)
                            skip_nodes.append(transport._node)
                            raise BadResource(e)
                        else:
                            raise
            except BadResource as e:
                if retry < (retry_count - 1):
                    continue
                else:
                    # Re-raise the inner exception
                    raise e.args[0]

    def _choose_pool(self, protocol=None):
        """
        Selects a connection pool according to the default protocol
        and the passed one.

        :param protocol: the protocol to use
        :type protocol: string
        :rtype: Pool
        """
        if not protocol:
            protocol = self.protocol
        if protocol == 'http':
            pool = self._http_pool
        elif protocol == 'pbc':
            pool = self._pb_pool
        else:
            raise ValueError("invalid protocol %s" % protocol)
        return pool


def _is_retryable(error):
    """
    Determines whether a given error is retryable according to the
    exceptions allowed to be retried by each transport.

    :param error: the error to check
    :type error: Exception
    :rtype: boolean
    """
    return is_pbc_retryable(error) or is_http_retryable(error)


def retryable(fn, protocol=None):
    """
    Wraps a client operation that can be retried according to the set
    :attr:`RiakClient.retries`. Used internally.
    """
    def wrapper(self, *args, **kwargs):
        pool = self._choose_pool(protocol)

        def thunk(transport):
            return fn(self, transport, *args, **kwargs)

        return self._with_retries(pool, thunk)

    wrapper.__doc__ = fn.__doc__
    wrapper.__repr__ = fn.__repr__

    return wrapper


def retryableHttpOnly(fn):
    """
    Wraps a retryable client operation that is only valid over HTTP.
    Used internally.
    """
    return retryable(fn, protocol='http')
